from random import Random
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ddapm_test_agent.trace import Span
from ddapm_test_agent.trace import Trace
from ddapm_test_agent.trace import dfs_order
from ddapm_test_agent.trace import root_span
from ddapm_test_agent.trace import verify_span


# Fix the seed for deterministic results between runs
_random = Random(1234)


_TYPES = {
    "web": {
        "name": ["django.request", "flask.request", "rack.request"],
        "service": ["django", "flask", "rack"],
        "resource": [
            "/users",
            "/feed",
            "/posts",
            "/home",
        ],
    },
    "cache": {
        "name": ["memcached.command"],
        "service": ["memcached"],
        "resource": [
            "get",
            "get_many",
            "set",
        ],
    },
    "sql": {
        "name": ["mysql.query", "postgres.query"],
        "service": ["mysql", "postgres"],
        "resource": [
            "SELECT * from users",
        ],
    },
}


def span(rnd: Random = _random, **kwargs: Any) -> Span:
    if "type" not in kwargs:
        kwargs["type"] = rnd.choice(list(_TYPES.keys()))

    data = _TYPES[kwargs["type"]]
    for k in ["name", "service"]:
        i = rnd.choice(list(range(len(data["name"]))))
        if k not in kwargs:
            kwargs[k] = data[k][i]

    if "resource" not in kwargs:
        kwargs["resource"] = rnd.choice(data["resource"])

    for k in ["trace_id", "span_id"]:
        if k not in kwargs:
            kwargs[k] = random_id(rnd)

    # Don't assign a parent id by default
    if "parent_id" not in kwargs:
        kwargs["parent_id"] = None

    if "start" not in kwargs:
        kwargs["start"] = rnd.randint(0, 2**48)

    if "duration" not in kwargs:
        kwargs["duration"] = rnd.randint(100, 10**10)

    if "meta" not in kwargs:
        kwargs["meta"] = {}

    if "metrics" not in kwargs:
        kwargs["metrics"] = {}
    return verify_span(kwargs)


def _prufers_trace(n: int, rnd: Random = _random) -> Trace:
    """Return a randomly generated trace tree with `n` spans.

    https://en.wikipedia.org/wiki/Pr%C3%BCfer_sequence
    """
    a = [rnd.randint(1, n) for _ in range(n - 2)]
    spans = [span(span_id=i) for i in range(1, n + 1)]
    degree: Dict[int, int] = {}
    for s in spans:
        degree[s["span_id"]] = 1
    for i in a:
        degree[i] += 1

    for i in a:
        for s in spans:
            if degree[s["span_id"]] == 1:
                s["parent_id"] = i
                degree[i] -= 1
                degree[s["span_id"]] -= 1
                break

    u = 0
    for s in spans:
        if degree[s["span_id"]] == 1:
            if u == 0:
                u = s["span_id"]
            else:
                s["parent_id"] = u
                break

    # TODO: randomize the span ids?
    return list(dfs_order(spans))


def random_trace(
    nspans: int,
    rng: Random = _random,
    trace_id: Optional[int] = None,
    parent_id: Optional[int] = None,
    remove_keys: Optional[List[str]] = None,
) -> Trace:
    # TODO:
    #   represent arbitrary random services (subtrees in spans)
    #   resource names (should only be on service entry)
    #   smarter type (should only be on service entry)
    #   sampling decisions
    #   dd_origin?
    assert nspans > 0
    if not trace_id:
        trace_id = random_id(rng)
    t = _prufers_trace(nspans, rng)
    root = root_span(t)
    if parent_id:
        root["parent_id"] = parent_id
    for s in t:
        if s is not root:
            del s["type"]
            del s["resource"]
        s["trace_id"] = trace_id
        if remove_keys:
            for k in remove_keys:
                del s[k]  # type: ignore

    return t


def random_id(rng: Random = _random) -> int:
    return rng.randint(0, 2**64)



import asyncio
import json
import os
from pathlib import Path
import random
import socket
import subprocess
from typing import AsyncGenerator
from typing import Awaitable
from typing import Dict
from typing import Generator
from typing import List
from typing import Literal
from typing import Optional
from typing import Set
from typing import cast

import aiohttp
from aiohttp.client_exceptions import ClientConnectorError
from aiohttp.client_exceptions import ClientOSError
from aiohttp.web import Response
from ddsketch import LogCollapsingLowestDenseDDSketch
from ddsketch.pb.proto import DDSketchProto
import msgpack
import pytest

from ddapm_test_agent.agent import _parse_csv
from ddapm_test_agent.agent import make_app
from ddapm_test_agent.apmtelemetry import TelemetryEvent
from ddapm_test_agent.trace import Span
from ddapm_test_agent.trace import Trace
from ddapm_test_agent.trace_snapshot import DEFAULT_SNAPSHOT_IGNORES


pytest_plugins = "aiohttp.pytest_plugin"


@pytest.fixture
def agent_enabled_checks() -> Generator[List[str], None, None]:
    yield ["trace_content_length", "meta_tracer_version_header", "trace_count_header", "trace_stall"]


@pytest.fixture
def log_span_fmt() -> Generator[str, None, None]:
    yield "[{name}]"


@pytest.fixture
def snapshot_dir(tmp_path: Path) -> Generator[Path, None, None]:
    yield tmp_path


@pytest.fixture
def snapshot_ci_mode() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def snapshot_ignored_attrs() -> Generator[Set[str], None, None]:
    yield set(_parse_csv(DEFAULT_SNAPSHOT_IGNORES))


@pytest.fixture
def agent_url() -> Generator[str, None, None]:
    yield ""


@pytest.fixture
def trace_request_delay() -> Generator[float, None, None]:
    yield 0.0


@pytest.fixture
def suppress_trace_parse_errors() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def pool_trace_check_failures() -> Generator[bool, None, None]:
    yield True


@pytest.fixture
def disable_error_responses() -> Generator[bool, None, None]:
    yield False


@pytest.fixture
def snapshot_removed_attrs() -> Generator[Set[str], None, None]:
    yield set()


@pytest.fixture
async def agent_app(
    aiohttp_server,
    agent_enabled_checks,
    log_span_fmt,
    snapshot_dir,
    snapshot_ci_mode,
    snapshot_ignored_attrs,
    agent_url,
    trace_request_delay,
    suppress_trace_parse_errors,
    pool_trace_check_failures,
    disable_error_responses,
    snapshot_removed_attrs,
):
    app = await aiohttp_server(
        make_app(
            agent_enabled_checks,
            log_span_fmt,
            str(snapshot_dir),
            snapshot_ci_mode,
            snapshot_ignored_attrs,
            agent_url,
            trace_request_delay,
            suppress_trace_parse_errors,
            pool_trace_check_failures,
            disable_error_responses,
            snapshot_removed_attrs,
        )
    )
    yield app


@pytest.fixture
async def agent(agent_app, aiohttp_client, loop):
    client = await aiohttp_client(agent_app)
    yield client


@pytest.fixture
def v04_reference_http_trace_payload_data_raw() -> List[Trace]:
    data = [
        [
            Span(
                {
                    "name": "http.request",
                    "service": "my-http-server",
                    "trace_id": random.randint(0, 2**64),
                    "span_id": random.randint(0, 2**64),
                    "parent_id": None,
                    "resource": "/users/",
                    "type": "http",
                    "start": 1342343123,
                    "duration": 123214,
                    "error": 0,
                    "meta": {
                        "component": "",
                        "http.url": "http://localhost:8080/users",
                        "http.method": "GET",
                        "http.status_code": "200",
                        "http.status_msg": "OK",
                        "language": "python",
                        "runtime-id": "2d377516ca12429aaf072f037ed2e4cc",
                    },
                    "metrics": {
                        "process_id": 111,
                        "sampling_priority_v1": 1.0,
                    },
                    "span_events": [
                        {
                            "name": "event1",
                            "time_unix_nano": 123,
                            "attributes": {
                                "string": {"type": 0, "string_value": "foo"},
                                "array": {
                                    "type": 4,
                                    "array_value": [{"type": 2, "int_value": 1}, {"type": 2, "int_value": 2}],
                                },
                            },
                        },
                    ],
                }
            ),
        ]
    ]
    return data


@pytest.fixture
def v04_reference_http_trace_payload_data(
    v04_reference_http_trace_payload_data_raw: List[Trace],
) -> bytes:
    return cast(bytes, msgpack.packb(v04_reference_http_trace_payload_data_raw))


@pytest.fixture
def v04_reference_http_trace_payload_headers() -> Dict[str, str]:
    headers = {
        "Content-Type": "application/msgpack",
        "X-Datadog-Trace-Count": "1",
        "Datadog-Meta-Tracer-Version": "v0.1",
        "datadog-meta-lang": "python",
        "X-Datadog-Trace-Env-Variables": "DD_INTEGRATION=express,DD_INTEGRATION_VERSION=1.2.3",
    }
    return headers


def v04_trace(  # type: ignore
    agent,
    traces: List[Trace],
    encoding: Literal["msgpack", "json"] = "msgpack",
    token: Optional[str] = None,
):
    params = {"test_session_token": token} if token is not None else {}
    if encoding == "msgpack":
        content_type = "application/msgpack"
        encode = msgpack.packb
    else:
        content_type = "application/json"
        encode = json.dumps

    headers = {
        "Content-Type": content_type,
        "X-Datadog-Trace-Count": str(len(traces)),
        "Datadog-Meta-Tracer-Version": "v0.1",
    }

    return agent.put(
        "/v0.4/traces",
        params=params,
        headers=headers,
        data=encode(traces),
    )


@pytest.fixture
def do_reference_v04_http_trace(
    agent,
    v04_reference_http_trace_payload_headers,
    v04_reference_http_trace_payload_data,
):
    def fn(token: Optional[str] = None, payload_override: Optional[bytes] = None) -> Awaitable[Response]:
        params = {"test_session_token": token} if token is not None else {}
        return agent.put(  # type: ignore
            "/v0.4/traces",
            params=params,
            headers=v04_reference_http_trace_payload_headers,
            data=payload_override or v04_reference_http_trace_payload_data,
        )

    yield fn


@pytest.fixture
def v06_reference_http_stats_payload_headers():
    headers = {
        "Content-Type": "application/msgpack",
        "Datadog-Meta-Lang": "python",
        "Datadog-Meta-Tracer-Version": "v0.1",
    }
    yield headers


@pytest.fixture
def v06_reference_http_stats_payload_data_raw():
    ok_dist = LogCollapsingLowestDenseDDSketch(0.00775, bin_limit=2048)
    err_dist = LogCollapsingLowestDenseDDSketch(0.00775, bin_limit=2048)

    rng = random.Random(0)
    total = 0
    ok_n = 97
    err_n = 3
    for _ in range(ok_n):
        n = rng.randint(int(1e9), int(2e9))
        total += n
        ok_dist.add(n)
    for _ in range(err_n):
        n = rng.randint(int(2e9), int(3e9))
        total += n
        err_dist.add(n)

    data = {
        "Env": "dev",
        "Version": "v0.1",
        "Hostname": "Host-1234",
        "Stats": [
            {
                "Start": 0,
                "Duration": 100000,
                "Stats": [
                    {
                        "Name": "http.request",
                        "Resource": "/user/profile",
                        "Service": "web-svc",
                        "Synthetics": False,
                        "Hits": ok_n + err_n,
                        "TopLevelHits": ok_n + err_n,
                        "Duration": total,
                        "Errors": err_n,
                        "OkSummary": DDSketchProto.to_proto(ok_dist).SerializeToString(),
                        "ErrorSummary": DDSketchProto.to_proto(err_dist).SerializeToString(),
                    },
                ],
            }
        ],
    }
    yield data


@pytest.fixture
def v06_reference_http_stats_payload_data(v06_reference_http_stats_payload_data_raw):
    yield msgpack.packb(v06_reference_http_stats_payload_data_raw)


@pytest.fixture
def do_reference_v06_http_stats(
    agent,
    v06_reference_http_stats_payload_headers,
    v06_reference_http_stats_payload_data,
):
    def fn(token: Optional[str] = None) -> Awaitable[Response]:
        params = {"test_session_token": token} if token is not None else {}
        return agent.put(  # type: ignore
            "/v0.6/stats",
            params=params,
            headers=v06_reference_http_stats_payload_headers,
            data=v06_reference_http_stats_payload_data,
        )

    yield fn


@pytest.fixture
def v2_reference_http_apmtelemetry_payload_data_raw():
    data = {
        "tracer_time": 1658439039,
        "runtime_id": "3cac6e9599564813977aace04bf37d57",
        "api_version": "v1",
        "seq_id": 1,
        "application": {
            "service_name": "my-svc",
            "service_version": "1.0.0",
            "env": "prod",
            "language_name": "python",
            "language_version": "3.9.10",
            "tracer_version": "1.3.0",
            "runtime_name": "CPython",
            "runtime_version": "3.9.10",
        },
        "host": {
            "os": "macOS-12.4",
            "hostname": "HELLO-COMPUTER",
            "os_version": "12.4",
            "kernel_name": "Darwin",
            "kernel_release": "21.5.0",
            "kernel_version": "Darwin Kernel Version 21.5.0: Tue Apr 26 21:08:22 PDT 2022; root:xnu-8020.121.3~4/RELEASE_X86_64",
            "container_id": "",
        },
        "payload": {
            "dependencies": [
                {"name": "pyparsing", "version": "3.0.9"},
                {"name": "pytest-mock", "version": "3.8.2"},
                {"name": "setuptools", "version": "62.6.0"},
                {"name": "sortedcontainers", "version": "2.4.0"},
                {"name": "attrs", "version": "21.4.0"},
                {"name": "wheel", "version": "0.37.1"},
                {"name": "protobuf", "version": "4.21.2"},
                {"name": "packaging", "version": "21.3"},
                {"name": "tomli", "version": "2.0.1"},
                {"name": "msgpack", "version": "1.0.4"},
                {"name": "bytecode", "version": "0.13.0"},
                {"name": "pip", "version": "22.1.2"},
                {"name": "py", "version": "1.11.0"},
                {"name": "ddsketch", "version": "2.0.3"},
                {"name": "coverage", "version": "6.4.2"},
                {"name": "pytest-cov", "version": "3.0.0"},
                {"name": "iniconfig", "version": "1.1.1"},
                {"name": "py-cpuinfo", "version": "8.0.0"},
                {"name": "toml", "version": "0.10.2"},
                {"name": "pluggy", "version": "1.0.0"},
                {"name": "mock", "version": "4.0.3"},
                {"name": "six", "version": "1.16.0"},
                {"name": "opentracing", "version": "2.4.0"},
                {"name": "pytest", "version": "6.2.5"},
                {"name": "ddtrace", "version": "1.3.0"},
                {"name": "tenacity", "version": "8.0.1"},
                {"name": "hypothesis", "version": "6.45.0"},
            ],
            "integrations": [],
            "configurations": [],
        },
        "request_type": "app-started",
    }
    yield data


@pytest.fixture
def v2_reference_http_apmtelemetry_payload_data(
    v2_reference_http_apmtelemetry_payload_data_raw,
):
    yield json.dumps(v2_reference_http_apmtelemetry_payload_data_raw)


@pytest.fixture
def v07_reference_http_remoteconfig_payload_data_raw():
    data = {
        "client": {
            "id": "test-id",
            "products": ["ASM_DATA", "ASM", "ASM_DD"],
            "is_tracer": True,
            "client_tracer": {
                "runtime_id": "test_runtime_id",
                "language": "python",
                "tracer_version": "test-1.x",
                "service": "test_agent_service",
                "env": "staging",
                "app_version": "0.1",
                "tags": ["env:staging", "version:0.1", "tracer_version:test-1.x", "host_name:test_agent_service"],
            },
            "state": {
                "root_version": 1,
                "targets_version": 46924638,
                "config_states": [
                    {"id": "blocking", "version": 296, "product": "ASM", "apply_state": 2},
                    {"id": "custom_rules", "version": 24, "product": "ASM", "apply_state": 2},
                    {"id": "disabled_rules", "version": 1, "product": "ASM", "apply_state": 2},
                    {"id": "exclusion_filters", "version": 256, "product": "ASM", "apply_state": 2},
                    {"id": "blocked_ips", "version": 4732, "product": "ASM_DATA", "apply_state": 2},
                    {"id": "blocked_users", "version": 5, "product": "ASM_DATA", "apply_state": 2},
                    {"id": "16.recommended.json", "version": 1, "product": "ASM_DD", "apply_state": 2},
                ],
                "has_error": False,
                "backend_client_state": "backend_client_state_hash",
            },
            "capabilities": "Afw=",
        },
        "cached_target_files": [
            {
                "path": "datadog/2/ASM/blocking/config",
                "length": 21,
                "hashes": [
                    {"algorithm": "sha256", "hash": "d75be3984bbd00451d1d56eff464d63a05727f192f0fa367d1aac03869b5d74c"}
                ],
            },
            {
                "path": "datadog/2/ASM/custom_rules/config",
                "length": 425,
                "hashes": [
                    {"algorithm": "sha256", "hash": "5c49e86c48d88e3e836e2cb85469ebf05c02a5d1bc768639e4c45c75620deb23"}
                ],
            },
            {
                "path": "datadog/2/ASM/disabled_rules/config",
                "length": 57,
                "hashes": [
                    {"algorithm": "sha256", "hash": "ab06f434cf8b9bc50bf75c27abada0781dfb2590aea9c19963b86e87e236fe1b"}
                ],
            },
            {
                "path": "datadog/2/ASM/exclusion_filters/config",
                "length": 21,
                "hashes": [
                    {"algorithm": "sha256", "hash": "d75be3984bbd00451d1d56eff464d63a05727f192f0fa367d1aac03869b5d74c"}
                ],
            },
            {
                "path": "datadog/2/ASM_DATA/blocked_ips/config",
                "length": 1290,
                "hashes": [
                    {"algorithm": "sha256", "hash": "a5e3b56034a6ad61d9853cd1b80061d04068ada2d5e3b1b120a68c861006db27"}
                ],
            },
            {
                "path": "datadog/2/ASM_DATA/blocked_users/config",
                "length": 760,
                "hashes": [
                    {"algorithm": "sha256", "hash": "53e2dd0da7a7d6ca57f10c0e1d17fdfc7e44a987bb128265258b3bc7d0c48193"}
                ],
            },
            {
                "path": "employee/ASM_DD/16.recommended.json/config",
                "length": 203383,
                "hashes": [
                    {"algorithm": "sha256", "hash": "ad984f3d9020ee932783ee6448cc4efc8a58aa2c8d3f4fd8d2d730452a8dd4df"}
                ],
            },
        ],
    }
    yield data


@pytest.fixture
def v07_reference_http_remoteconfig_payload_data(
    v07_reference_http_remoteconfig_payload_data_raw,
):
    yield json.dumps(v07_reference_http_remoteconfig_payload_data_raw)


@pytest.fixture
def v2_reference_http_apmtelemetry_payload_headers(  # type: ignore
    v2_reference_http_apmtelemetry_payload_data_raw,
) -> Generator[Dict[str, str], None, None]:
    headers = {
        "Content-type": "application/json",
        "DD-Telemetry-Request-Type": v2_reference_http_apmtelemetry_payload_data_raw["request_type"],
        "DD-Telemetry-API-Version": "v1",
    }
    yield headers


def v2_apmtelemetry(  # type: ignore
    agent,
    event: TelemetryEvent,
    token: Optional[str] = None,
):
    params = {"test_session_token": token} if token is not None else {}
    headers = {
        "Content-type": "application/json",
        "DD-Telemetry-Request-Type": event["request_type"],
        "DD-Telemetry-API-Version": "v1",
    }

    return agent.post(
        "/telemetry/proxy/api/v2/apmtelemetry",
        params=params,
        headers=headers,
        data=json.dumps(event),
    )


@pytest.fixture
def do_reference_v2_http_apmtelemetry(
    agent,
    v2_reference_http_apmtelemetry_payload_headers,
    v2_reference_http_apmtelemetry_payload_data,
):
    def fn(token: Optional[str] = None) -> Awaitable[Response]:
        params = {"test_session_token": token} if token is not None else {}
        return agent.post(  # type: ignore
            "/telemetry/proxy/api/v2/apmtelemetry",
            params=params,
            headers=v2_reference_http_apmtelemetry_payload_headers,
            data=v2_reference_http_apmtelemetry_payload_data,
        )

    yield fn


@pytest.fixture
def available_port() -> str:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))  # Bind to a free port provided by the host.
    port = s.getsockname()[1]  # Get the port number assigned.
    s.close()  # Release the socket.
    return str(port)


@pytest.fixture
def testagent_port(available_port: str) -> str:
    return available_port


@pytest.fixture
def testagent_url(testagent_port: str) -> str:
    return "http://localhost:%s" % testagent_port


@pytest.fixture(scope="module")
def testagent_snapshot_ci_mode() -> bool:
    # Default all tests in this module to be run in CI mode
    # unless a special env var is passed to make generating
    # the snapshots easier.
    return os.getenv("GENERATE_SNAPSHOTS") != "1"


@pytest.fixture
async def testagent(
    loop: asyncio.BaseEventLoop, testagent_port: str, testagent_snapshot_ci_mode: bool
) -> AsyncGenerator[aiohttp.ClientSession, None]:
    env = os.environ.copy()
    env.update(
        {
            "PORT": testagent_port,
            "SNAPSHOT_CI": "1" if testagent_snapshot_ci_mode else "0",
            "SNAPSHOT_DIR": os.path.join(os.path.dirname(__file__), "integration_snapshots"),
        }
    )
    p = subprocess.Popen(["ddapm-test-agent"], env=env)

    # Wait for server to start
    try:
        async with aiohttp.ClientSession() as session:
            for _ in range(100):
                try:
                    r = await session.get(f"http://localhost:{testagent_port}")
                except (ClientConnectorError, ClientOSError):
                    pass
                else:
                    if r.status == 404:
                        break
                await asyncio.sleep(0.05)
            else:
                assert 0
            yield session
    finally:
        p.terminate()



import base64
import datetime
import hashlib
import json
from typing import Any
from typing import Dict
from typing import Optional


class RemoteConfigServer:
    _responses: Dict[Optional[str], Any] = {}

    def _update_response(self, token: Optional[str], data: Dict[str, Any]) -> None:
        if self._responses.get(token):
            self._responses[token].update(data)
        else:
            self._create_response(token, data)

    def _create_response(self, token: Optional[str], data: Dict[str, Any]) -> None:
        self._responses[token] = data

    async def _get_response(self, token: Optional[str]) -> Any:
        return self._responses.get(token, {})

    def update_config_response(self, token: Optional[str], data: Dict[str, Any]) -> None:
        self._update_response(token, data)

    @staticmethod
    def _build_config_path_response(path: str, msg: str) -> Dict[str, Any]:
        expires_date = datetime.datetime.strftime(
            datetime.datetime.now() + datetime.timedelta(days=1), "%Y-%m-%dT%H:%M:%SZ"
        )
        msg_enc = bytes(json.dumps(msg), encoding="utf-8")
        data = {
            "signatures": [{"keyid": "", "sig": ""}],
            "signed": {
                "_type": "targets",
                "custom": {"opaque_backend_state": ""},
                "expires": expires_date,
                "spec_version": "1.0.0",
                "targets": {
                    path: {
                        "custom": {"c": [""], "v": 0},
                        "hashes": {"sha256": hashlib.sha256(msg_enc).hexdigest()},
                        "length": len(msg_enc),
                    }
                },
                "version": 0,
            },
        }
        remote_config_payload = {
            "roots": [
                str(
                    base64.b64encode(
                        bytes(
                            json.dumps(
                                {
                                    "signatures": [],
                                    "signed": {
                                        "_type": "root",
                                        "consistent_snapshot": True,
                                        "expires": "1986-12-11T00:00:00Z",
                                        "keys": {},
                                        "roles": {},
                                        "spec_version": "1.0",
                                        "version": 2,
                                    },
                                }
                            ),
                            encoding="utf-8",
                        )
                    ),
                    encoding="utf-8",
                )
            ],
            "targets": str(base64.b64encode(bytes(json.dumps(data), encoding="utf-8")), encoding="utf-8"),
            "target_files": [
                {
                    "path": path,
                    "raw": str(base64.b64encode(msg_enc), encoding="utf-8"),
                }
            ],
            "client_configs": [path],
        }
        return remote_config_payload

    def create_config_path_response(self, token: Optional[str], path: str, msg: str) -> None:
        remote_config_payload = self._build_config_path_response(path, msg)
        self.create_config_response(token, remote_config_payload)

    def create_config_response(self, token: Optional[str], data: Dict[str, Any]) -> None:
        self._create_response(token, data)

    async def get_config_response(self, token: Optional[str]) -> Any:
        return await self._get_response(token)



import asyncio
import json
import logging
from typing import Any
from typing import Dict
from typing import cast

from aiohttp.web import Request


log = logging.getLogger(__name__)
TelemetryEvent = Dict[str, Any]


async def v2_decode_request(request: Request, data: bytes) -> TelemetryEvent:
    headers = request.headers

    if "X-Datadog-Test-Stall-Seconds" in headers:
        duration = float(headers["X-Datadog-Test-Stall-Seconds"])
    else:
        duration = request.app["trace_request_delay"]
    if duration > 0:
        log.info("Stalling for %r seconds.", duration)
        await asyncio.sleep(duration)
    return v2_decode(data)


def v2_decode(data: bytes) -> TelemetryEvent:
    """Decode v2 apm telemetry request data as a dict"""
    # TODO: Handle decoding into a telemetry payload object
    return cast(TelemetryEvent, json.loads(data))



class Integration:
    def __init__(self, integration_name: str, integration_version: str, dependency_name: str):
        self.integration_name = integration_name
        self.integration_version = integration_version
        self.dependency_name = dependency_name



import json
import os

import pytest

from ddapm_test_agent import trace_snapshot
from ddapm_test_agent import tracestats_snapshot
from ddapm_test_agent.trace import add_span_event
from ddapm_test_agent.trace import add_span_link
from ddapm_test_agent.trace import copy_span
from ddapm_test_agent.trace import set_attr
from ddapm_test_agent.trace import set_meta_tag
from ddapm_test_agent.trace import set_metric_tag
from ddapm_test_agent.tracestats import StatsAggr
from ddapm_test_agent.tracestats import StatsBucket

from .conftest import v04_trace
from .trace_utils import random_trace


@pytest.mark.parametrize("snapshot_ci_mode", [False, True])
async def test_snapshot_single_trace(
    agent,
    snapshot_dir,
    snapshot_ci_mode,
    do_reference_v04_http_trace,
):
    """
    When a trace is sent and a snapshot taken
        When not in CI mode
            The test should fail
        When in CI mode
            The snapshot file should be created
            When the same trace is sent again
                The snapshot should pass
    """  # noqa: RST301
    # Send a trace
    resp = await do_reference_v04_http_trace(token="test_case")
    assert resp.status == 200

    # Do the snapshot
    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test_case"})
    snap_path = snapshot_dir / "test_case.json"
    if snapshot_ci_mode:
        # No previous snapshot file exists so this should fail
        assert resp.status == 400, await resp.text()
        assert f"Trace snapshot file '{snap_path}' not found" in await resp.text()
    else:
        # Since this is the first invocation the snapshot file should be created
        assert resp.status == 200, await resp.text()
        assert os.path.exists(snap_path)
        with open(snap_path, mode="r") as f:
            assert "".join(f.readlines()) != ""

        # Do the snapshot again to actually perform a comparison
        resp = await do_reference_v04_http_trace(token="test_case")
        assert resp.status == 200, await resp.text()

        resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test_case"})
        assert resp.status == 200, await resp.text()


ONE_SPAN_TRACE = random_trace(1)
TWO_SPAN_TRACE = random_trace(2)
FIVE_SPAN_TRACE = random_trace(5)


@pytest.mark.parametrize(
    "expected_traces,actual_traces,error",
    [
        ([ONE_SPAN_TRACE], [ONE_SPAN_TRACE], ""),
        ([FIVE_SPAN_TRACE], [FIVE_SPAN_TRACE], ""),
        # Mismatching trace sizes
        (
            [TWO_SPAN_TRACE],
            [TWO_SPAN_TRACE[:-1]],
            "Received fewer spans (1) than expected (2). Expected unmatched spans: 'postgres.query'",
        ),
        (
            [TWO_SPAN_TRACE[:-1]],
            [TWO_SPAN_TRACE],
            "Received more spans (2) than expected (1). Received unmatched spans: 'postgres.query'",
        ),
        (
            [[set_attr(copy_span(ONE_SPAN_TRACE[0]), "name", "name_expected")]],
            [[set_attr(copy_span(ONE_SPAN_TRACE[0]), "name", "name_received")]],
            "span mismatch on 'name': got 'name_received' which does not match expected 'name_expected'",
        ),
        (
            [
                [
                    TWO_SPAN_TRACE[0],
                    set_attr(copy_span(TWO_SPAN_TRACE[1]), "name", "name_expected"),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE[0],
                    set_attr(copy_span(TWO_SPAN_TRACE[1]), "name", "name_received"),
                ]
            ],
            "span mismatch on 'name': got 'name_received' which does not match expected 'name_expected'",
        ),
        (
            [
                [
                    TWO_SPAN_TRACE[0],
                    set_meta_tag(copy_span(TWO_SPAN_TRACE[1]), "expected", "value"),
                ]
            ],
            [[TWO_SPAN_TRACE[0], TWO_SPAN_TRACE[1]]],
            "Span meta value 'expected' in expected span but is not in the received span.",
        ),
        (
            [[TWO_SPAN_TRACE[0], TWO_SPAN_TRACE[1]]],
            [
                [
                    TWO_SPAN_TRACE[0],
                    set_metric_tag(copy_span(TWO_SPAN_TRACE[1]), "received", 123.32),
                ]
            ],
            "Span metrics value 'received' in received span but is not in the expected span.",
        ),
        # Mismatching metrics tag
        (
            [
                [
                    TWO_SPAN_TRACE[0],
                    set_metric_tag(copy_span(TWO_SPAN_TRACE[1]), "received", 123.32),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE[0],
                    set_metric_tag(copy_span(TWO_SPAN_TRACE[1]), "received", 123.32),
                ]
            ],
            "",
        ),
        # Default ignored fields
        (
            [
                [
                    {
                        "name": "s",
                        "span_id": 1234,
                        "trace_id": 1,
                        "parent_id": 0,
                        "resource": "/",
                        "start": 0,
                        "duration": 1,
                        "type": "web",
                        "error": 0,
                        "meta": {},
                        "metrics": {},
                    }
                ]
            ],
            [
                [
                    {
                        "name": "s",
                        "span_id": 4321,
                        "trace_id": 2,
                        "parent_id": 0,
                        "resource": "/",
                        "start": 0,
                        "duration": 1,
                        "type": "web",
                        "error": 0,
                        "meta": {},
                        "metrics": {},
                    }
                ]
            ],
            "",
        ),
    ],
)
async def test_snapshot_trace_differences(agent, expected_traces, actual_traces, error):
    resp = await v04_trace(agent, expected_traces, token="test")
    assert resp.status == 200, await resp.text()

    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test"})
    assert resp.status == 200, await resp.text()
    resp = await agent.get("/test/session/clear", params={"test_session_token": "test"})
    assert resp.status == 200, await resp.text()

    resp = await v04_trace(agent, actual_traces, token="test")
    assert resp.status == 200, await resp.text()
    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test"})
    resp_text = await resp.text()
    if error:
        assert resp.status == 400, resp_text
        assert error in resp_text, resp_text
    else:
        assert resp.status == 200, resp_text


@pytest.mark.parametrize(
    "trace,expected",
    [
        (
            [
                [
                    {"trace_id": 1, "span_id": 1, "start": 0},
                    {"trace_id": 1, "parent_id": 1, "span_id": 2, "start": 1},
                    {"trace_id": 1, "parent_id": 1, "span_id": 3, "start": 2},
                    {"trace_id": 1, "parent_id": 2, "span_id": 4, "start": 4},
                ]
            ],
            """[[
  {
    "trace_id": 0,
    "span_id": 1,
    "parent_id": 0,
    "start": 0
  },
     {
       "trace_id": 0,
       "span_id": 2,
       "parent_id": 1,
       "start": 1
     },
        {
          "trace_id": 0,
          "span_id": 4,
          "parent_id": 2,
          "start": 4
        },
     {
       "trace_id": 0,
       "span_id": 3,
       "parent_id": 1,
       "start": 2
     }]]\n""",
        ),
        (
            [
                [
                    {"trace_id": 1, "parent_id": None, "span_id": 1, "start": 0},
                    {"trace_id": 1, "parent_id": 1, "span_id": 2, "start": 1},
                    {"trace_id": 1, "parent_id": 1, "span_id": 3, "start": 2},
                    {"trace_id": 1, "parent_id": 2, "span_id": 4, "start": 4},
                ]
            ],
            """[[
  {
    "trace_id": 0,
    "span_id": 1,
    "parent_id": 0,
    "start": 0
  },
     {
       "trace_id": 0,
       "span_id": 2,
       "parent_id": 1,
       "start": 1
     },
        {
          "trace_id": 0,
          "span_id": 4,
          "parent_id": 2,
          "start": 4
        },
     {
       "trace_id": 0,
       "span_id": 3,
       "parent_id": 1,
       "start": 2
     }]]\n""",
        ),
    ],
)
def test_generate_trace_snapshot(trace, expected):
    assert trace_snapshot.generate_snapshot(trace) == expected


@pytest.mark.parametrize(
    "buckets,expected",
    [
        (
            [
                StatsBucket(  # noqa
                    Start=1000,
                    Duration=10,
                    Stats=[
                        # Not using all the fields of StatsAggr, hence the ignores
                        StatsAggr(  # type: ignore
                            Name="http.request",
                            Type="http",
                            Resource="/users/list",
                            Hits=10,
                            TopLevelHits=10,
                            Duration=100,
                        ),  # noqa
                        StatsAggr(  # type: ignore
                            Name="http.request",
                            Type="http",
                            Resource="/users/create",
                            Hits=5,
                            TopLevelHits=5,
                            Duration=10,
                        ),  # noqa
                    ],
                ),
                StatsBucket(
                    Start=1010,
                    Duration=10,
                    Stats=[
                        StatsAggr(  # type: ignore
                            Name="http.request",
                            Type="http",
                            Resource="/users/list",
                            Hits=20,
                            TopLevelHits=20,
                            Duration=200,
                        ),
                    ],
                ),
            ],
            """[
  {
    "Start": 0,
    "Duration": 10,
    "Stats": [
      {
        "Name": "http.request",
        "Type": "http",
        "Resource": "/users/create",
        "Hits": 5,
        "TopLevelHits": 5,
        "Duration": 10
      },
      {
        "Name": "http.request",
        "Type": "http",
        "Resource": "/users/list",
        "Hits": 10,
        "TopLevelHits": 10,
        "Duration": 100
      }
    ]
  },
  {
    "Start": 10,
    "Duration": 10,
    "Stats": [
      {
        "Name": "http.request",
        "Type": "http",
        "Resource": "/users/list",
        "Hits": 20,
        "TopLevelHits": 20,
        "Duration": 200
      }
    ]
  }
]\n""",
        ),
    ],
)
def test_generate_tracestats_snapshot(buckets, expected):
    assert tracestats_snapshot.generate(buckets) == expected


async def test_snapshot_custom_dir(agent, tmp_path, do_reference_v04_http_trace):
    resp = await do_reference_v04_http_trace(token="test_case")
    assert resp.status == 200

    custom_dir = tmp_path / "custom"
    custom_dir.mkdir()

    resp = await agent.get(
        "/test/session/snapshot",
        params={"test_session_token": "test_case", "dir": str(custom_dir)},
    )
    snap_path = custom_dir / "test_case.json"
    assert resp.status == 200, await resp.text()
    assert os.path.exists(snap_path)
    with open(snap_path, mode="r") as f:
        assert "".join(f.readlines()) != ""


async def test_snapshot_custom_file(agent, tmp_path, do_reference_v04_http_trace):
    resp = await do_reference_v04_http_trace(token="test_case")
    assert resp.status == 200

    custom_dir = tmp_path / "custom"
    custom_dir.mkdir()
    custom_file_name = custom_dir / "custom_snapshot"
    custom_file = custom_dir / "custom_snapshot.json"

    resp = await agent.get(
        "/test/session/snapshot",
        params={"test_session_token": "test_case", "file": str(custom_file_name)},
    )
    assert resp.status == 200, await resp.text()
    assert os.path.exists(custom_file), custom_file
    with open(custom_file, mode="r") as f:
        assert "".join(f.readlines()) != ""


@pytest.mark.parametrize("snapshot_ci_mode", [False, True])
async def test_snapshot_tracestats(agent, tmp_path, snapshot_ci_mode, do_reference_v06_http_stats, snapshot_dir):
    resp = await do_reference_v06_http_stats(token="test_case")
    assert resp.status == 200

    snap_path = snapshot_dir / "test_case_tracestats.json"
    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test_case"})
    resp_clear = await agent.get("/test/session/clear", params={"test_session_token": "test_case"})
    assert resp_clear.status == 200, await resp_clear.text()

    if snapshot_ci_mode:
        # No previous snapshot file exists so this should fail
        assert resp.status == 400
        assert f"Trace stats snapshot file '{snap_path}' not found" in await resp.text()
    else:
        # First invocation the snapshot, file should be created
        assert resp.status == 200
        assert os.path.exists(snap_path)
        with open(snap_path, mode="r") as f:
            assert "".join(f.readlines()) != ""

        # Do the snapshot again to actually perform a comparison
        resp = await do_reference_v06_http_stats(token="test_case")
        assert resp.status == 200, await resp.text()

        resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test_case"})
        assert resp.status == 200, await resp.text()


@pytest.mark.parametrize("snapshot_removed_attrs", [{"start", "duration", "span_events.name"}])
async def test_removed_attributes(agent, tmp_path, snapshot_removed_attrs, do_reference_v04_http_trace):
    resp = await do_reference_v04_http_trace(token="test_case")
    assert resp.status == 200

    custom_dir = tmp_path / "custom"
    custom_dir.mkdir()
    custom_file_name = custom_dir / "custom_snapshot"
    custom_file = custom_dir / "custom_snapshot.json"

    resp = await agent.get(
        "/test/session/snapshot", params={"test_session_token": "test_case", "file": str(custom_file_name)}
    )
    assert resp.status == 200, await resp.text()

    assert os.path.exists(custom_file), custom_file
    with open(custom_file, mode="r") as f:  # Check that the removed attributes are not present in the span
        file_content = "".join(f.readlines())
        assert file_content != ""
        span = json.loads(file_content)
        for removed_attr in snapshot_removed_attrs:
            assert removed_attr not in span[0]


@pytest.mark.parametrize("snapshot_removed_attrs", [{"metrics.process_id"}])
async def test_removed_attributes_metrics(agent, tmp_path, snapshot_removed_attrs, do_reference_v04_http_trace):
    resp = await do_reference_v04_http_trace(token="test_case")
    assert resp.status == 200

    custom_dir = tmp_path / "custom"
    custom_dir.mkdir()
    custom_file_name = custom_dir / "custom_snapshot"
    custom_file = custom_dir / "custom_snapshot.json"

    resp = await agent.get(
        "/test/session/snapshot", params={"test_session_token": "test_case", "file": str(custom_file_name)}
    )
    assert resp.status == 200, await resp.text()

    assert os.path.exists(custom_file), custom_file
    with open(custom_file, mode="r") as f:
        file_content = "".join(f.readlines())
        assert file_content != ""
        span = json.loads(file_content)
        assert "process_id" not in span[0]


ONE_SPAN_TRACE_NO_START = random_trace(1, remove_keys=["start"])
TWO_SPAN_TRACE_NO_START = random_trace(2, remove_keys=["start"])
FIVE_SPAN_TRACE_NO_START = random_trace(5, remove_keys=["start"])


@pytest.mark.parametrize(
    "expected_traces,actual_traces,error,snapshot_removed_attrs",
    [
        ([ONE_SPAN_TRACE_NO_START], [ONE_SPAN_TRACE_NO_START], "", {"start"}),
        ([FIVE_SPAN_TRACE_NO_START], [FIVE_SPAN_TRACE_NO_START], "", {"start"}),
        # Mismatching trace sizes
        (
            [TWO_SPAN_TRACE_NO_START],
            [TWO_SPAN_TRACE_NO_START[:-1]],
            "Received fewer spans (1) than expected (2). Expected unmatched spans: 'flask.request'",
            {"start"},
        ),
        (
            [TWO_SPAN_TRACE_NO_START[:-1]],
            [TWO_SPAN_TRACE_NO_START],
            "Received more spans (2) than expected (1). Received unmatched spans: 'flask.request'",
            {"start"},
        ),
        (
            [[set_attr(copy_span(ONE_SPAN_TRACE_NO_START[0]), "name", "name_expected")]],
            [[set_attr(copy_span(ONE_SPAN_TRACE_NO_START[0]), "name", "name_received")]],
            "span mismatch on 'name': got 'name_received' which does not match expected 'name_expected'",
            {"start"},
        ),
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    set_attr(copy_span(TWO_SPAN_TRACE_NO_START[1]), "name", "name_expected"),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    set_attr(copy_span(TWO_SPAN_TRACE_NO_START[1]), "name", "name_received"),
                ]
            ],
            "span mismatch on 'name': got 'name_received' which does not match expected 'name_expected'",
            {"start"},
        ),
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    set_meta_tag(copy_span(TWO_SPAN_TRACE_NO_START[1]), "expected", "value"),
                ]
            ],
            [[TWO_SPAN_TRACE_NO_START[0], TWO_SPAN_TRACE_NO_START[1]]],
            "Span meta value 'expected' in expected span but is not in the received span.",
            {"start"},
        ),
        (
            [[TWO_SPAN_TRACE_NO_START[0], TWO_SPAN_TRACE_NO_START[1]]],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    set_metric_tag(copy_span(TWO_SPAN_TRACE_NO_START[1]), "received", 123.32),
                ]
            ],
            "Span metrics value 'received' in received span but is not in the expected span.",
            {"start"},
        ),
        # Mismatching metrics tag
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    set_metric_tag(copy_span(TWO_SPAN_TRACE_NO_START[1]), "received", 123.32),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    set_metric_tag(copy_span(TWO_SPAN_TRACE_NO_START[1]), "received", 123.32),
                ]
            ],
            "",
            {"start"},
        ),
        # Mismatching span links count
        (
            [
                TWO_SPAN_TRACE_NO_START,
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0]),
                ]
            ],
            "Span value 'span_links' in received span but is not in the expected span.",
            {"start"},
        ),
        # Mismatching span link reference
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[1]),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0]),
                ]
            ],
            "Span link 0 mismatch on 'span_id': got '1' which does not match expected '2'.",
            {"start"},
        ),
        # Mismatching span link value
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], flags=1),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], flags=0),
                ]
            ],
            "Span link 0 mismatch on 'flags': got '0' which does not match expected '1'.",
            {"start"},
        ),
        # Mismatching span link fields
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0]),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], flags=1),
                ]
            ],
            "Span link 0 value 'flags' in received span link but is not in the expected span link.",
            {"start"},
        ),
        # Mismatching span link attribute
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(
                        copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], {"a": "2", "b": "3"}
                    ),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(
                        copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], {"a": "2", "b": "0"}
                    ),
                ]
            ],
            "Span link 0 attributes mismatch on 'b': got '0' which does not match expected '3'.",
            {"start"},
        ),
        # Matching span link
        (
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(
                        copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], {"a": "2", "b": "0"}, 1
                    ),
                ]
            ],
            [
                [
                    TWO_SPAN_TRACE_NO_START[0],
                    add_span_link(
                        copy_span(TWO_SPAN_TRACE_NO_START[1]), TWO_SPAN_TRACE_NO_START[0], {"a": "2", "b": "0"}, 1
                    ),
                ]
            ],
            "",
            {"start"},
        ),
        # Mismatching span events count
        (
            [TWO_SPAN_TRACE_NO_START],
            [[TWO_SPAN_TRACE_NO_START[0], add_span_event(copy_span(TWO_SPAN_TRACE_NO_START[1]))]],
            "Span value 'span_events' in received span but is not in the expected span.",
            {"start"},
        ),
        # Mismatching span event name
        (
            [[add_span_event(copy_span(ONE_SPAN_TRACE_NO_START[0]), name="expected_name")]],
            [[add_span_event(copy_span(ONE_SPAN_TRACE_NO_START[0]), name="got_name")]],
            "Span event 0 mismatch on 'name': got 'got_name' which does not match expected 'expected_name'.",
            {"start"},
        ),
        # Mismatching span event attributes
        (
            [[add_span_event(copy_span(ONE_SPAN_TRACE_NO_START[0]), attributes={"a": "1", "b": "2"})]],
            [[add_span_event(copy_span(ONE_SPAN_TRACE_NO_START[0]), attributes={"a": "1", "b": "0"})]],
            "Span event 0 attributes mismatch on 'b': got '{'type': 0, 'string_value': '0'}' which does not match expected '{'type': 0, 'string_value': '2'}'.",
            {"start"},
        ),
        # Matching span event
        (
            [[add_span_event(copy_span(ONE_SPAN_TRACE_NO_START[0]), attributes={"a": "1", "b": 2, "c": [3]})]],
            [[add_span_event(copy_span(ONE_SPAN_TRACE_NO_START[0]), attributes={"a": "1", "b": 2, "c": [3]})]],
            "",
            {"start"},
        ),
        # Default ignored fields
        (
            [
                [
                    {
                        "name": "s",
                        "span_id": 1234,
                        "trace_id": 1,
                        "parent_id": 0,
                        "resource": "/",
                        "duration": 1,
                        "type": "web",
                        "error": 0,
                        "meta": {},
                        "metrics": {},
                        "span_events": [
                            {
                                "time_unix_nano": 123,
                                "name": "event_name",
                            },
                        ],
                    }
                ]
            ],
            [
                [
                    {
                        "name": "s",
                        "span_id": 4321,
                        "trace_id": 2,
                        "parent_id": 0,
                        "resource": "/",
                        "duration": 1,
                        "type": "web",
                        "error": 0,
                        "meta": {},
                        "metrics": {},
                        "span_events": [
                            {
                                "time_unix_nano": 456,
                                "name": "event_name",
                            },
                        ],
                    }
                ]
            ],
            "",
            {"start"},
        ),
    ],
)
async def test_snapshot_trace_differences_removed_start(agent, expected_traces, actual_traces, error):
    resp = await v04_trace(agent, expected_traces, token="test")
    assert resp.status == 200, await resp.text()

    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test"})
    assert resp.status == 200, await resp.text()
    resp = await agent.get("/test/session/clear", params={"test_session_token": "test"})
    assert resp.status == 200, await resp.text()

    resp = await v04_trace(agent, actual_traces, token="test")
    assert resp.status == 200, await resp.text()
    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test"})
    resp_text = await resp.text()
    if error:
        assert resp.status == 400, resp_text
        assert error in resp_text, resp_text
    else:
        assert resp.status == 200, resp_text


@pytest.mark.parametrize("snapshot_removed_attrs", [{"span_id"}])
async def test_removed_attributes_fails_span_id(agent, tmp_path, snapshot_removed_attrs, do_reference_v04_http_trace):
    resp = await do_reference_v04_http_trace(token="test_case")
    assert resp.status == 200, await resp.text()

    resp = await agent.get("/test/session/snapshot", params={"test_session_token": "test_case"})
    assert resp.status == 400
    assert "Cannot remove 'span_id' from spans" in await resp.text()



from typing import List
from typing import Optional
from typing import TypedDict

from ddsketch import DDSketch
from ddsketch.pb.ddsketch_pb2 import DDSketch as DDSketchProto
import msgpack


# Note that class attributes are golang style to match the payload.


class StatsAggr(TypedDict):
    Name: str
    Resource: str
    Service: Optional[str]
    Type: Optional[str]
    HTTPStatusCode: int
    Synthetics: bool
    Hits: int
    TopLevelHits: int
    Duration: int
    Errors: int
    OkSummary: DDSketch
    ErrorSummary: DDSketch


class StatsBucket(TypedDict):
    Start: int
    Duration: int
    Stats: List[StatsAggr]


class v06StatsPayload(TypedDict):
    Hostname: Optional[str]
    Env: Optional[str]
    Version: Optional[str]
    Stats: List[StatsBucket]


def decode_v06(data: bytes) -> v06StatsPayload:
    payload = msgpack.unpackb(data)
    stats_buckets: List[StatsBucket] = []
    for raw_bucket in payload["Stats"]:
        stats: List[StatsAggr] = []
        for raw_stats in raw_bucket["Stats"]:
            stat = StatsAggr(
                Name=raw_stats["Name"],
                Resource=raw_stats["Resource"],
                Service=raw_stats.get("Service"),
                Type=raw_stats.get("Type"),
                HTTPStatusCode=raw_stats.get("HTTPStatusCode"),
                Synthetics=raw_stats["Synthetics"],
                Hits=raw_stats["Hits"],
                TopLevelHits=raw_stats["TopLevelHits"],
                Duration=raw_stats["Duration"],
                Errors=raw_stats["Errors"],
                OkSummary=DDSketchProto().ParseFromString(raw_stats["OkSummary"]),
                ErrorSummary=DDSketchProto().ParseFromString(raw_stats["ErrorSummary"]),
            )
            stats.append(stat)

        bucket = StatsBucket(
            Start=raw_bucket["Start"],
            Duration=raw_bucket["Duration"],
            Stats=stats,
        )
        stats_buckets.append(bucket)

    return v06StatsPayload(
        Hostname=payload.get("Hostname"),
        Env=payload.get("Env"),
        Version=payload.get("Version"),
        Stats=stats_buckets,
    )



from collections import OrderedDict
import json
import logging
import operator
import pprint
import textwrap
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import OrderedDict as OrderedDictType
from typing import Set
from typing import Tuple
from typing import cast

from .checks import Check
from .checks import CheckTrace
from .trace import Span
from .trace import SpanId
from .trace import TopLevelSpanValue
from .trace import Trace
from .trace import TraceId
from .trace import bfs_order
from .trace import child_map
from .trace import copy_trace
from .trace import root_span
from .trace import trace_id as get_trace_id


log = logging.getLogger(__name__)


DEFAULT_SNAPSHOT_IGNORES = "span_id,trace_id,parent_id,duration,start,metrics.system.pid,metrics.system.process_id,metrics.process_id,metrics._dd.tracer_kr,meta.runtime-id,span_links.trace_id_high,span_events.time_unix_nano,meta.pathway.hash,meta._dd.p.tid"


def _key_match(d1: Dict[str, Any], d2: Dict[str, Any], key: str) -> bool:
    """
    >>> _key_match({"a": 1}, {"a": 2}, "a")
    False
    >>> _key_match({"a": 1}, {"a": 2}, "b")
    True
    >>> _key_match({"a": 2}, {"a": 1}, "a")
    False
    >>> _key_match({"a": 1}, {"a": 1}, "a")
    True
    >>> _key_match({"a": 2}, {"b": 1}, "a")
    False
    >>> _key_match({"b": 2}, {"a": 1}, "a")
    False
    """
    try:
        return (key not in d1 and key not in d2) or cast(bool, d1[key] == d2[key])
    except KeyError:
        return False


def _span_similarity(s1: Span, s2: Span) -> int:
    score = 0

    for key in ["name", "service", "type", "error", "resource"]:
        if not _key_match(s1, s2, key):  # type: ignore
            score -= 1

    s1_meta = s1.get("meta", {})
    s2_meta = s2.get("meta", {})
    for key in set(s1_meta.keys()) | set(s2_meta.keys()):
        if not _key_match(s1_meta, s2_meta, key):
            score -= 1

    s1_metrics = s1.get("metrics", {})
    s2_metrics = s2.get("metrics", {})
    for key in set(s1_metrics.keys()) | set(s2_metrics.keys()):
        if not _key_match(s1_metrics, s2_metrics, key):
            score -= 1
    return score


def _trace_similarity(t1: Trace, t2: Trace) -> int:
    score = 0
    score -= abs(len(t1) - len(t2))
    score += _span_similarity(root_span(t1), root_span(t2))
    # TODO: also compare child spans? try to match them somehow?
    return score


def _normalize_traces(traces: List[Trace]) -> List[Trace]:
    normed_traces = []
    trace_id_map: Dict[TraceId, Tuple[int, Dict[SpanId, int]]] = {}

    for trace_id, trace in enumerate(traces):
        normed_trace = copy_trace(list(bfs_order(trace)))
        # Have to start at 1 since 0 is reserved for indicating no parent (i.e. root span)
        span_id = 1
        old_trace_id = 0

        span_id_map: Dict[SpanId, int] = {}
        for span in normed_trace:
            old_trace_id = span["trace_id"]
            span["trace_id"] = trace_id
            span_id_map[span["span_id"]] = span_id
            span["span_id"] = span_id
            parent_id = span.get("parent_id")
            if parent_id:
                # If parent_id is not in the map, assume this is a trace chunk with
                # a parent not in the trace chunk. Eg: distributed traces.
                span["parent_id"] = span_id_map.get(parent_id, parent_id)
            else:
                # Normalize the parent of root spans to be 0.
                span["parent_id"] = 0

            if "meta" not in span:
                span["meta"] = {}
            if "metrics" not in span:
                span["metrics"] = {}
            span_id += 1

        normed_traces.append(normed_trace)

        if old_trace_id != 0:
            trace_id_map[old_trace_id] = (trace_id, span_id_map)

    for normed_trace in normed_traces:
        for span in normed_trace:
            if "span_links" in span:
                for link in span["span_links"]:
                    if link["trace_id"] in trace_id_map:
                        (trace_id, span_id_map) = trace_id_map[link["trace_id"]]
                        link["trace_id"] = trace_id
                        if link["span_id"] in span_id_map:
                            link["span_id"] = span_id_map[link["span_id"]]
                    # TODO: maybe normalize traceparent too?

    return normed_traces


def _match_traces(t1s: List[Trace], t2s: List[Trace]) -> List[Tuple[Trace, Trace]]:
    t1_map: Dict[TraceId, Trace] = {get_trace_id(t): t for t in t1s}
    t2_map: Dict[TraceId, Trace] = {get_trace_id(t): t for t in t2s}
    # Note: cannot assume trace_ids are unique between t1s and t2s (since snapshot normalizes them)
    similarities: List[Tuple[TraceId, TraceId, int]] = []

    # Map the similarity of every trace in t1s to every trace in t2s
    for t1_trace_id, t1 in t1_map.items():
        for t2_trace_id, t2 in t2_map.items():
            similarities.append((t1_trace_id, t2_trace_id, _trace_similarity(t1, t2)))

    # Sort the similarities by score
    similarities = sorted(similarities, key=lambda x: x[2], reverse=True)

    # Go over the scores and match the traces
    matched_t1s = set()
    matched_t2s = set()
    matches: List[Tuple[Trace, Trace]] = []
    for t1_trace_id, t2_trace_id, score in similarities:
        if (t1_trace_id in matched_t1s) or (t2_trace_id in matched_t2s):
            continue
        matches.append((t1_map[t1_trace_id], t2_map[t2_trace_id]))
        matched_t1s.add(t1_trace_id)
        matched_t2s.add(t2_trace_id)

    if len(matched_t1s) != len(t1_map):
        unmatched_ids = set(t1_map.keys()) - matched_t1s
        op_names = [t1_map[tid][0]["name"] for tid in unmatched_ids]
        raise AssertionError("Did not receive expected traces: %s" % (",".join(map(lambda s: f"'{s}'", op_names))))
    if len(matched_t2s) != len(t2_map):
        unmatched_ids = set(t2_map.keys()) - matched_t2s
        op_names = [t2_map[tid][0]["name"] for tid in unmatched_ids]
        raise AssertionError("Received unmatched traces: %s" % (",".join(map(lambda s: f"'{s}'", op_names))))
    return matches


def _diff_spans(
    s1: Span, s2: Span, ignored: Set[str]
) -> Tuple[List[str], List[str], List[str], List[Tuple[List[str], List[str]]], List[Tuple[List[str], List[str]]]]:
    """Return differing attributes between two spans and their meta/metrics maps.

    It is assumed that the spans have passed through preliminary validation
    to ensure all required fields are included.

    >>> from .trace import verify_span, copy_span
    >>> span = verify_span(dict(name="", trace_id=1234, span_id=11, meta={}, metrics={}))
    >>> span2 = copy_span(span)
    >>> span2["resource"] = ""
    >>> _diff_spans(span, span2, set())
    (['resource'], [], [], [], [])
    >>> span2["type"] = "web"
    >>> tuple(map(set, _diff_spans(span, span2, set()))) == ({'resource', 'type'}, set(), set(), set(), set())
    True
    >>> span2["meta"]["key"] = "value"
    >>> tuple(map(set, _diff_spans(span, span2, set()))) == ({'resource', 'type'}, {'key'}, set(), set(), set())
    True
    >>> span2["metrics"]["key2"] = 100.0
    >>> tuple(map(set, _diff_spans(span, span2, set()))) == ({'resource', 'type'}, {'key'}, {'key2'}, set(), set())
    True
    >>> _diff_spans(span, span2, set(['metrics.key2', 'meta.key', 'resource', 'type', 'meta.key2']))
    ([], [], [], [], [])
    """
    results = []
    s1_no_tags = cast(
        Dict[str, TopLevelSpanValue],
        {k: v for k, v in s1.items() if k not in ("meta", "metrics", "span_links", "span_events")},
    )
    s2_no_tags = cast(
        Dict[str, TopLevelSpanValue],
        {k: v for k, v in s2.items() if k not in ("meta", "metrics", "span_links", "span_events")},
    )
    for d1, d2, ign in [
        (s1_no_tags, s2_no_tags, ignored),
        (s1["meta"], s2["meta"], set(i[5:] for i in ignored if i.startswith("meta."))),
        (
            s1["metrics"],
            s2["metrics"],
            set(i[8:] for i in ignored if i.startswith("metrics.")),
        ),
    ]:
        d1 = cast(Dict[str, Any], d1)
        d2 = cast(Dict[str, Any], d2)
        diffs = []
        for k in (set(d1.keys()) | set(d2.keys())) - ign:
            if not _key_match(d1, d2, k):
                diffs.append(k)
        results.append(diffs)

    link_diffs = []
    if len(s1.get("span_links") or []) != len(s2.get("span_links") or []):
        results[0].append("span_links")
    else:
        for l1, l2 in zip(s1.get("span_links") or [], s2.get("span_links") or []):
            l1_no_tags = cast(
                Dict[str, TopLevelSpanValue],
                {k: v for k, v in l1.items() if k != "attributes"},
            )
            l2_no_tags = cast(
                Dict[str, TopLevelSpanValue],
                {k: v for k, v in l2.items() if k != "attributes"},
            )
            link_diff = []
            for d1, d2, ign in [
                (l1_no_tags, l2_no_tags, set(i[11:] for i in ignored if i.startswith("span_links."))),
                (
                    l1.get("attributes") or {},
                    l2.get("attributes") or {},
                    set(i[22:] for i in ignored if i.startswith("span_links.attributes.")),
                ),
            ]:
                d1 = cast(Dict[str, Any], d1)
                d2 = cast(Dict[str, Any], d2)
                diffs = []
                for k in (set(d1.keys()) | set(d2.keys())) - ign:
                    if not _key_match(d1, d2, k):
                        diffs.append(k)
                link_diff.append(diffs)

            link_diffs.append(link_diff)
    results.append(link_diffs)  # type: ignore

    event_diffs = []
    if len(s1.get("span_events") or []) != len(s2.get("span_events") or []):
        results[0].append("span_events")
    else:
        for e1, e2 in zip(s1.get("span_events") or [], s2.get("span_events") or []):
            l1_no_tags = cast(
                Dict[str, TopLevelSpanValue],
                {k: v for k, v in e1.items() if k != "attributes"},
            )
            l2_no_tags = cast(
                Dict[str, TopLevelSpanValue],
                {k: v for k, v in e2.items() if k != "attributes"},
            )
            event_diff = []
            for d1, d2, ign in [
                (l1_no_tags, l2_no_tags, set(i[12:] for i in ignored if i.startswith("span_events."))),
                (
                    e1.get("attributes") or {},
                    e2.get("attributes") or {},
                    set(i[23:] for i in ignored if i.startswith("span_events.attributes.")),
                ),
            ]:
                diffs = []
                d1 = cast(Dict[str, Any], d1)
                d2 = cast(Dict[str, Any], d2)
                for k in (set(d1.keys()) | set(d2.keys())) - ign:
                    if not _key_match(d1, d2, k):
                        diffs.append(k)
                event_diff.append(diffs)

            event_diffs.append(event_diff)
    results.append(event_diffs)  # type: ignore

    return cast(
        Tuple[List[str], List[str], List[str], List[Tuple[List[str], List[str]]], List[Tuple[List[str], List[str]]]],
        tuple(results),
    )


def _compare_traces(expected: Trace, received: Trace, ignored: Set[str]) -> None:
    """Compare two traces for differences.

    The given traces are assumed to be in BFS order.
    """
    if len(received) > len(expected):
        names = ["'%s'" % s["name"] for s in received[len(expected) - len(received) :]]
        raise AssertionError(
            f"Received more spans ({len(received)}) than expected ({len(expected)}). Received unmatched spans: {', '.join(names)}"
        )
    elif len(expected) > len(received):
        names = ["'%s'" % s["name"] for s in expected[len(received) - len(expected) :]]
        raise AssertionError(
            f"Received fewer spans ({len(received)}) than expected ({len(expected)}). Expected unmatched spans: {', '.join(names)}"
        )

    for s_exp, s_rec in zip(expected, received):
        with CheckTrace.add_frame(
            f"snapshot compare of span '{s_exp['name']}' at position {s_exp['span_id']} in trace"
        ) as frame:
            frame.add_item(f"Expected span:\n{pprint.pformat(s_exp)}")
            frame.add_item(f"Received span:\n{pprint.pformat(s_rec)}")
            top_level_diffs, meta_diffs, metrics_diffs, span_link_diffs, span_event_diffs = _diff_spans(
                s_exp, s_rec, ignored
            )

            for diffs, diff_type, d_exp, d_rec in [
                (top_level_diffs, "span", s_exp, s_rec),
                (meta_diffs, "meta", s_exp["meta"], s_rec["meta"]),
                (metrics_diffs, "metrics", s_exp["metrics"], s_rec["metrics"]),
            ]:
                for diff_key in diffs:
                    if diff_key not in d_exp:
                        raise AssertionError(
                            f"Span{' ' + diff_type if diff_type != 'span' else ''} value '{diff_key}' in received span but is not in the expected span."
                        )
                    elif diff_key not in d_rec:
                        raise AssertionError(
                            f"Span{' ' + diff_type if diff_type != 'span' else ''} value '{diff_key}' in expected span but is not in the received span."
                        )
                    elif diff_key in ["span_links", "span_events"]:
                        raise AssertionError(
                            f"{diff_type} mismatch on '{diff_key}': got {len(d_rec[diff_key])} values for {diff_key} which does not match expected {len(d_exp[diff_key])}."  # type: ignore
                        )
                    else:
                        raise AssertionError(
                            f"{diff_type} mismatch on '{diff_key}': got '{d_rec[diff_key]}' which does not match expected '{d_exp[diff_key]}'."
                        )

            for i, (link_level_diffs, attribute_diffs) in enumerate(span_link_diffs):
                for diffs, diff_type, d_exp, d_rec in [
                    (link_level_diffs, f"{i}", s_exp["span_links"][i], s_rec["span_links"][i]),
                    (
                        attribute_diffs,
                        f"{i} attributes",
                        s_exp["span_links"][i].get("attributes") or {},
                        s_rec["span_links"][i].get("attributes") or {},
                    ),
                ]:
                    for diff_key in diffs:
                        if diff_key not in d_exp:
                            raise AssertionError(
                                f"Span link {diff_type} value '{diff_key}' in received span link but is not in the expected span link."
                            )
                        elif diff_key not in d_rec:
                            raise AssertionError(
                                f"Span link {diff_type} value '{diff_key}' in expected span link but is not in the received span link."
                            )
                        else:
                            raise AssertionError(
                                f"Span link {diff_type} mismatch on '{diff_key}': got '{d_rec[diff_key]}' which does not match expected '{d_exp[diff_key]}'."
                            )

            for i, (event_level_diffs, attribute_diffs) in enumerate(span_event_diffs):
                for diffs, diff_type, d_exp, d_rec in [
                    (event_level_diffs, f"{i}", s_exp["span_events"][i], s_rec["span_events"][i]),
                    (
                        attribute_diffs,
                        f"{i} attributes",
                        s_exp["span_events"][i].get("attributes") or {},
                        s_rec["span_events"][i].get("attributes") or {},
                    ),
                ]:
                    for diff_key in diffs:
                        if diff_key not in d_exp:
                            raise AssertionError(
                                f"Span event {diff_type} value '{diff_key}' in received span event but is not in the expected span event."
                            )
                        elif diff_key not in d_rec:
                            raise AssertionError(
                                f"Span event {diff_type} value '{diff_key}' in expected span event but is not in the received span event."
                            )
                        else:
                            raise AssertionError(
                                f"Span event {diff_type} mismatch on '{diff_key}': got '{d_rec[diff_key]}' which does not match expected '{d_exp[diff_key]}'."
                            )


class SnapshotFailure(Exception):
    pass


class SnapshotCheck(Check):
    def check(self, *args, **kwargs):
        pass


def snapshot(expected_traces: List[Trace], received_traces: List[Trace], ignored: List[str]) -> None:
    normed_expected = _normalize_traces(expected_traces)
    normed_received = _normalize_traces(received_traces)
    with CheckTrace.add_frame(
        f"compare of {len(normed_expected)} expected trace(s) to {len(normed_received)} received trace(s)"
    ):
        matched = _match_traces(normed_expected, normed_received)
        log.debug("Matched traces %r", matched)

        for exp, rec in matched:
            with CheckTrace.add_frame(f"trace '{exp[0]['name']}' ({len(exp)} spans)"):
                _compare_traces(exp, rec, set(ignored))


def _ordered_span(s: Span) -> OrderedDictType[str, TopLevelSpanValue]:
    """Order the span to be more human-readable."""
    d = OrderedDict()
    order = [
        "name",
        "service",
        "resource",
        "trace_id",
        "span_id",
        "parent_id",
        "type",
        "error",
        "meta",
        "metrics",
        "span_links",
        "span_events",
    ]
    for k in order:
        if k in s:
            if k in ["meta", "metrics"]:
                # Sort the meta and metrics dictionaries alphanumerically
                d[k] = OrderedDict(sorted(s[k].items(), key=operator.itemgetter(0)))  # type: ignore
            else:
                d[k] = s[k]  # type: ignore

    # Add the rest of the attributes in alphanumeric order.
    for k in sorted(set(s.keys()) - set(order)):
        d[k] = s[k]  # type: ignore

    if "span_links" in d:
        for link in d["span_links"]:
            if "attributes" in link:
                link["attributes"] = OrderedDict(sorted(link["attributes"].items(), key=operator.itemgetter(0)))

    if "span_events" in d:
        for event in d["span_events"]:
            if "attributes" in event:
                event["attributes"] = OrderedDict(sorted(event["attributes"].items(), key=operator.itemgetter(0)))

    for k in ["meta", "metrics"]:
        if k in d and len(d[k]) == 0:
            del d[k]
    return d  # type: ignore


def _snapshot_trace_str(trace: Trace, removed: Optional[List[str]] = None) -> str:
    cmap = child_map(trace)
    stack: List[Tuple[int, Span]] = [(0, root_span(trace))]
    s = "[\n"
    while stack:
        prefix, span = stack.pop(0)

        # Remove any keys that are not needed for comparison
        if removed:
            for key in removed:
                if key.startswith("meta."):
                    span["meta"].pop(key[5:], None)
                elif key.startswith("metrics."):
                    span["metrics"].pop(key[8:], None)
                elif key.startswith("span_links.attributes."):
                    if "span_links" in span:
                        for link in span["span_links"]:
                            if "attributes" in link:
                                link["attributes"].pop(key[22:], None)
                elif key.startswith("span_links."):
                    if "span_links" in span:
                        for link in span["span_links"]:
                            link.pop(key[11:], None)  # type: ignore
                elif key.startswith("span_events.attributes."):
                    if "span_events" in span:
                        for event in span["span_events"]:
                            if "attributes" in event:
                                event["attributes"].pop(key[23:], None)
                elif key.startswith("span_events."):
                    if "span_events" in span:
                        for event in span["span_events"]:
                            event.pop(key[12:], None)  # type: ignore
                else:
                    span.pop(key, None)  # type: ignore

        for i, child in enumerate(reversed(cmap[span["span_id"]])):
            if i == 0:
                stack.insert(0, (prefix + 3, child))
            else:
                stack.insert(0, (prefix + 3, child))

        s += textwrap.indent(json.dumps(_ordered_span(span), indent=2), " " * (prefix + 2))
        if stack:
            s += ",\n"
    s += "]"
    return s


def _snapshot_json(traces: List[Trace], removed: Optional[List[str]] = None) -> str:
    s = "["
    for t in traces:
        s += _snapshot_trace_str(t, removed)
        if t != traces[-1]:
            s += ",\n"
    s += "]\n"
    return s


def generate_snapshot(received_traces: List[Trace], removed: Optional[List[str]] = None) -> str:
    return _snapshot_json(_normalize_traces(received_traces), removed)



from ddtrace import Tracer
import pytest

from ddapm_test_agent.client import TestAgentClient


@pytest.fixture
async def client(testagent, testagent_url):
    return TestAgentClient(testagent_url)


@pytest.fixture
async def tracer(testagent_url):
    t = Tracer(testagent_url)
    yield t
    t.shutdown()


async def test_client_traces(client: TestAgentClient, tracer: Tracer) -> None:
    assert client.traces() == []
    with tracer.trace("test"):
        pass
    traces = client.wait_for_num_traces(1)
    assert len(traces) == 1


async def test_client_requests(client: TestAgentClient, tracer: Tracer) -> None:
    assert client.requests() == []
    with tracer.trace("test"):
        pass
    tracer.flush()
    assert len(client.requests()) == 1



import asyncio
import contextlib
import contextvars
import dataclasses
import textwrap
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple
from typing import Type


CHECK_TRACE: contextvars.ContextVar["CheckTrace"] = contextvars.ContextVar("check_trace")


class CheckNotFound(IndexError):
    pass


class CheckTraceFrame:
    def __init__(self, name: str) -> None:
        self._checks: List["Check"] = []
        self._name = name
        self._children: List["CheckTraceFrame"] = []
        self._items: List[str] = []

    def add_item(self, item: str) -> None:
        self._items.append(item)

    def add_check(self, check: "Check") -> None:
        self._checks.append(check)

    def add_frame(self, frame: "CheckTraceFrame") -> None:
        self._children.append(frame)

    def has_fails(self) -> bool:
        for c in self._checks:
            if c.failed:
                return True
        return False

    def update_results(self, results: Dict[str, Dict[str, int]]) -> Dict[str, Dict[str, int]]:
        """Return as follows
        output = { check.name: { "Passed_Checks": int, "Failed_Checks": int, "Skipped_Checks": int}}

        """
        for c in self._checks:
            if c.failed:
                results[c.name]["Failed_Checks"] += 1
            elif c.skipped:
                results[c.name]["Skipped_Checks"] += 1
            else:
                results[c.name]["Passed_Checks"] += 1
        return results

    def __repr__(self) -> str:
        return f"<CheckTraceFrame name='{self._name}' children={len(self._children)}>"


class CheckTrace:
    """A trace of a check used to provide helpful debugging information for failed checks."""

    def __init__(self, root: CheckTraceFrame):
        self._root = root
        self._active = root

    @classmethod
    @contextlib.contextmanager
    def add_frame(cls, title: str) -> Generator["CheckTraceFrame", None, None]:
        """Add a frame to the trace."""
        ctx = CHECK_TRACE.get()
        frame = CheckTraceFrame(title)
        ctx._active.add_frame(frame)
        prev_active = ctx._active
        ctx._active = frame
        yield frame
        ctx._active = prev_active

    @classmethod
    def add_check(cls, check: "Check") -> None:
        """Add the given check to the current frame."""
        ctx = CHECK_TRACE.get()
        ctx._active.add_check(check)

    def frames(self) -> Generator[CheckTraceFrame, None, None]:
        fs: List[CheckTraceFrame] = [self._root]
        while fs:
            frame = fs.pop(0)
            yield frame
            fs = frame._children + fs

    def frames_dfs(self) -> Generator[Tuple[CheckTraceFrame, int], None, None]:
        fs: List[Tuple[CheckTraceFrame, int]] = [(self._root, 0)]
        while fs:
            frame, depth = fs.pop(0)
            yield frame, depth
            fs = [(f, depth + 1) for f in frame._children] + fs

    def has_fails(self) -> bool:
        return len([f for f in self.frames() if f.has_fails()]) > 0

    def update_results(self, results: Dict[str, Dict[str, int]]) -> Dict[str, Dict[str, int]]:
        for f in self.frames():
            f.update_results(results)
        return results

    def get_failures_by_check(self, failures_by_check: Dict[str, List[str]]) -> Dict[str, List[str]]:
        # TODO?: refactor so this code isnt duplicated with __str__
        frame_s = ""
        for frame, depth in self.frames_dfs():
            indent = " " * (depth + 2) if depth > 0 else ""

            frame_s += f"{indent}At {frame._name}:\n"
            for item in frame._items:
                frame_s += textwrap.indent(f"- {item}", prefix=f" {indent}")
                frame_s += "\n"

            for c in frame._checks:
                if c.failed:
                    check_s = f"{indent}❌ Check '{c.name}' failed: {c._msg}\n"
                    if c.name in failures_by_check:
                        failures_by_check[c.name].append(frame_s + check_s)
                    else:
                        failures_by_check[c.name] = [frame_s + check_s]
        return failures_by_check

    def __str__(self) -> str:
        s = ""
        # TODO?: only include frames that have fails
        for frame, depth in self.frames_dfs():
            indent = " " * (depth + 2) if depth > 0 else ""
            s += f"{indent}At {frame._name}:\n"
            for item in frame._items:
                s += textwrap.indent(f"- {item}", prefix=f" {indent}")
                s += "\n"

            for c in frame._checks:
                if c.failed:
                    s += f"{indent}❌ Check '{c.name}' failed: {c._msg}\n"
        return s


class Check:
    name: str
    """Name of the check. Should be as succinct and representative as possible."""

    description: str
    """Description of the check. Be as descriptive as possible as this will be included
    with error messages returned to the test case.
    """

    def __init__(self) -> None:
        self._failed: bool = False
        self._skipped: bool = False
        self._msg: str = ""

    @property
    def failed(self) -> bool:
        return self._failed

    @property
    def skipped(self) -> bool:
        return self._skipped

    def fail(self, msg: str) -> None:
        self._failed = True
        self._msg = msg

    def skip(self, msg: str) -> None:
        self._skipped = True
        self._msg = msg

    def check(self, *args, **kwargs):
        """Perform any checking required for this Check.

        CheckFailures should be raised for any failing checks.
        """
        raise NotImplementedError


@dataclasses.dataclass()
class Checks:
    checks: List[Type[Check]] = dataclasses.field(init=True)
    enabled: List[str] = dataclasses.field(init=True)

    def _get_check(self, name: str) -> Type[Check]:
        for c in self.checks:
            if c.name == name:
                return c
        else:
            raise CheckNotFound("Check for code %r not found" % name)

    def is_enabled(self, name: str) -> bool:
        check = self._get_check(name)
        if check.name in self.enabled:
            return True
        return False

    async def check(self, name: str, *args: Any, **kwargs: Any) -> None:
        """Find and run the check with the given ``name`` if it is enabled."""
        check = self._get_check(name)()

        if self.is_enabled(name):
            # Register the check with the current trace
            CheckTrace.add_check(check)

            # Run the check
            if asyncio.iscoroutinefunction(check.check):
                await check.check(*args, **kwargs)
            else:
                check.check(*args, **kwargs)


def start_trace(msg: str) -> CheckTrace:
    trace = CheckTrace(CheckTraceFrame(name=msg))
    CHECK_TRACE.set(trace)
    return trace



import json
from typing import List

from .checks import CheckTrace
from .tracestats import StatsBucket


def _normalize_statsbuckets(buckets: List[StatsBucket]) -> List[StatsBucket]:
    """Normalize the stats bucket by normalizing the time buckets."""
    # Make a copy of the buckets, note that the sketches are not copied.
    normed_buckets = []
    for bucket in buckets:
        bcopy = bucket.copy()
        bcopy["Stats"] = [aggr.copy() for aggr in bucket["Stats"]]  # Copy the aggregations
        normed_buckets.append(bcopy)

    # Order the buckets by time
    normed_buckets = sorted(normed_buckets, key=lambda b: b["Start"])

    # Sort aggr for a bucket alphanumerically
    for bucket in normed_buckets:
        # Sort aggrs to get a consistent list of buckets, since order does not matter.
        bucket["Stats"] = sorted(
            bucket["Stats"],
            key=lambda b: (
                b["Name"],
                b["Resource"],
                b["Type"],
                b.get("Synthetics"),
                b.get("HTTPStatusCode"),
                b["Hits"],
            ),
        )

    start = normed_buckets[0]["Start"]
    for b in normed_buckets:
        b["Start"] -= start

    return normed_buckets


def snapshot(expected_stats: List[StatsBucket], received_stats: List[StatsBucket]) -> None:
    # Normalize the stats buckets by making them independent of time. Only ordering matters.

    # Sort the buckets by start time.
    normed_expected = _normalize_statsbuckets(expected_stats)
    normed_received = _normalize_statsbuckets(received_stats)

    # TODO: do better matching and comparing to aid in debugging
    assert len(normed_received) == len(
        normed_expected
    ), f"Number of stats buckets ({len(normed_received)}) doesn't match expected ({len(normed_expected)})."

    with CheckTrace.add_frame(f"snapshot compare of {len(normed_received)} stats buckets"):
        # Do a really rough comparison.
        for i, (exp_bucket, rec_bucket) in enumerate(zip(normed_expected, normed_received)):
            with CheckTrace.add_frame(f"comparison of bucket {i}") as frame:
                exp_aggrs = exp_bucket["Stats"]
                rec_aggrs = rec_bucket["Stats"]
                frame.add_item(f"Expected aggregated stats: {exp_aggrs}")
                frame.add_item(f"Received aggregated stats: {rec_aggrs}")
                assert len(exp_aggrs) == len(
                    rec_aggrs
                ), f"Number of aggregations ({len(rec_aggrs)}) in bucket {i} doesn't match expected: ({len(exp_aggrs)})."

                for j, (exp_aggr, rec_aggr) in enumerate(zip(exp_aggrs, rec_aggrs)):
                    # Omit duration and sketches for now
                    # Duration and sketches will be noisy
                    for attr in (
                        "Name",
                        "Resource",
                        "Type",
                        "Synthetics",
                        "Hits",
                        "TopLevelHits",
                        "Errors",
                        "HTTPStatusCode",
                    ):
                        exp_value, rec_value = exp_aggr[attr], rec_aggr[attr]  # type: ignore
                        if exp_value != rec_value:
                            raise AssertionError(
                                f"Expected value ('{exp_value}') for field '{attr}' does not match received value: '{rec_value}'."
                            )


def generate(received_stats: List[StatsBucket]) -> str:
    return f"{json.dumps(_normalize_statsbuckets(received_stats), indent=2)}\n"



"""Tracing specific functions and types"""

import json
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List
from typing import Literal
from typing import Optional
from typing import OrderedDict
from typing import Tuple
from typing import Union
from typing import cast

import msgpack
from typing_extensions import NotRequired
from typing_extensions import TypedDict


SpanId = int
TraceId = int


SPAN_TYPES = [
    "cache",
    "cassandra",
    "elasticsearch",
    "grpc",
    "http",
    "mongodb",
    "redis",
    "sql",
    "template",
    "test",
    "web",
    "worker",
]
SPAN_REQUIRED_ATTRS = [
    "name",
    "span_id",
    "trace_id",
    "duration",
    "start",
    "parent_id",
]

MetricType = Union[int, float]


class SpanLink(TypedDict):
    trace_id: int
    trace_id_high: int
    span_id: SpanId
    attributes: NotRequired[Dict[str, str]]
    tracestate: NotRequired[Optional[str]]
    flags: NotRequired[Optional[int]]


class SpanEvent(TypedDict):
    time_unix_nano: int
    name: str
    attributes: NotRequired[Dict[str, Dict[str, Any]]]


class Span(TypedDict):
    name: str
    span_id: SpanId
    trace_id: TraceId
    start: int
    duration: int
    parent_id: NotRequired[Optional[int]]
    service: NotRequired[Optional[str]]
    resource: NotRequired[Optional[str]]
    type: NotRequired[Optional[str]]  # noqa
    error: NotRequired[Optional[int]]
    meta: NotRequired[Dict[str, str]]
    metrics: NotRequired[Dict[str, MetricType]]
    span_links: NotRequired[List[SpanLink]]
    span_events: NotRequired[List[SpanEvent]]
    meta_struct: NotRequired[Dict[str, Dict[str, Any]]]


SpanAttr = Literal[
    "name",
    "span_id",
    "trace_id",
    "start",
    "duration",
    "parent_id",
    "service",
    "resource",
    "type",
    "error",
    "meta",
    "metrics",
    "span_links",
    "span_events",
    "meta_struct",
]
TopLevelSpanValue = Union[
    None,
    SpanId,
    TraceId,
    int,
    str,
    Dict[str, str],
    Dict[str, MetricType],
    List[SpanLink],
    List[SpanEvent],
]
Trace = List[Span]
v04TracePayload = List[List[Span]]
TraceMap = OrderedDict[int, Trace]


# TODO:ban add extra tags to add to the span
# TODO:ban warn about dropping metastruct
def verify_span(d: Any) -> Span:
    try:
        assert isinstance(d, dict), f"Expected 'span' to be of type: '{dict}', got: '{type(d)}'"
        # TODO: check these
        required_attrs = ["span_id", "trace_id", "name"]
        for attr in required_attrs:
            assert attr in d, f"'{attr}' required in span"
        NoneType = type(None)
        assert isinstance(d["span_id"], int), "Expected 'span_id' to be of type: 'int', got: " + str(type(d["span_id"]))
        assert isinstance(d["trace_id"], int), "Expected 'trace_id' to be of type: 'int', got: " + str(
            type(d["trace_id"])
        )
        assert isinstance(d["name"], str), "Expected 'name' to be of type: 'str', got: " + str(type(d["name"]))
        if "resource" in d:
            assert isinstance(d["resource"], (str, NoneType)), "Expected 'resource' to be of type: 'str', got: " + str(type(d["resource"]))  # type: ignore
        if "service" in d:
            assert isinstance(d["service"], (str, NoneType)), "Expected 'service' to be of type: 'str', got: " + str(type(d["service"]))  # type: ignore
        if "type" in d:
            assert isinstance(d["type"], (str, NoneType)), "Expected 'type' to be of type: 'str', got: " + str(type(d["type"]))  # type: ignore
        if "parent_id" in d:
            assert isinstance(d["parent_id"], (int, NoneType)), "Expected 'parent_id' to be of type: 'int', got: " + str(type(d["parent_id"]))  # type: ignore
        if "error" in d:
            assert isinstance(d["error"], int), "Expected error to be of type: 'int', got: " + str(type(d["error"]))
        if "meta" in d:
            assert isinstance(d["meta"], dict)
            for k, attr in d["meta"].items():
                assert isinstance(k, str), f"Expected key 'meta.{k}' to be of type: 'str', got: {type(k)}"
                assert isinstance(
                    attr, str
                ), f"Expected value of key 'meta.{k}' to be of type: 'str', got: {type(attr)}"
        if "meta_struct" in d:
            assert isinstance(d["meta_struct"], dict)
            for k, val in d["meta_struct"].items():
                assert isinstance(k, str), f"Expected key 'meta_struct.{k}' to be of type: 'str', got: {type(k)}"
                assert isinstance(
                    val, dict
                ), f"Expected msgpack decoded value of key 'meta_struct.{k}' to be of type: 'dict', got: {type(val)}"
                for inner_k in val:
                    assert isinstance(
                        inner_k, str
                    ), f"Expected key 'meta_struct.{k}.{inner_k}' to be of type: 'str', got: {type(inner_k)}"
        if "metrics" in d:
            assert isinstance(d["metrics"], dict)
            for k, attr in d["metrics"].items():
                assert isinstance(k, str), f"Expected key 'metrics.{k}' to be of type: 'str', got: {type(k)}"
                assert isinstance(
                    attr, (int, float)
                ), f"Expected value of key 'metrics.{k}' to be of type: 'float/int', got: {type(attr)}"
        if "span_links" in d:
            assert isinstance(d["span_links"], list)
            for link in d["span_links"]:
                assert isinstance(link, dict), f"Expected all span_links to be of type: 'dict', got: {type(link)}"
                required_attrs = ["span_id", "trace_id"]
                for attr in required_attrs:
                    assert attr in link, f"'{attr}' required in span link"
                assert isinstance(link["span_id"], int), "Expected 'span_id' to be of type: 'int', got: " + str(
                    type(link["span_id"])
                )
                assert isinstance(link["trace_id"], int), "Expected 'trace_id' to be of type: 'int', got: " + str(
                    type(link["trace_id"])
                )
                if "trace_id_high" in link:
                    assert isinstance(
                        link["trace_id_high"], (int, NoneType)  # type: ignore
                    ), "Expected 'trace_id_high' to be of type: 'int', got: " + str(type(link["trace_id_high"]))
                if "attributes" in link:
                    assert isinstance(link["attributes"], dict)
                    for k, attr in link["attributes"].items():
                        assert isinstance(k, str), f"Expected key 'attributes.{k}' to be of type: 'str', got: {type(k)}"
                        assert isinstance(
                            attr, str
                        ), f"Expected value of key 'attributes.{k}' to be of type: 'str', got: {type(attr)}"
                if "tracestate" in link:
                    assert isinstance(
                        link["tracestate"], (str, NoneType)  # type: ignore
                    ), "Expected 'tracestate' to be of type: 'str', got: " + str(type(link["tracestate"]))
                if "flags" in link:
                    assert isinstance(link["flags"], int), "Expected flags to be of type: 'int', got: " + str(
                        type(link["flags"])
                    )
        if "span_events" in d:
            assert isinstance(d["span_events"], list)
            for event in d["span_events"]:
                assert isinstance(event, dict), f"Expected all span_events to be of type: 'dict', got: {type(event)}"
                required_attrs = ["time_unix_nano", "name"]
                for attr in required_attrs:
                    assert attr in event, f"'{attr}' required in span event"
                assert isinstance(
                    event["time_unix_nano"], int
                ), "Expected 'time_unix_nano' to be of type: 'int', got: " + str(type(event["time_unix_nano"]))
                assert isinstance(event["name"], str), "Expected 'name' to be of type: 'str', got: " + str(
                    type(event["name"])
                )
                if "attributes" in event:
                    assert isinstance(event["attributes"], dict)
                    for k, attr in event["attributes"].items():
                        assert isinstance(k, str), f"Expected key 'attributes.{k}' to be of type: 'str', got: {type(k)}"
                        assert isinstance(
                            attr, dict
                        ), f"Expected value 'attributes.{k}={attr}' to be of type: 'dict', got: {type(attr)}"

                        # for this data struture for attr:
                        # // AttributeAnyValue is an Object with String keys, defined below.
                        # // We have to implement a union type manually here because Go's MessagePack generator does not support
                        # // unions: https://github.com/tinylib/msgp/issues/184
                        # {
                        #   // Represents the type of the value represented in this attribute entry.
                        #   // For String values: "type": 0
                        #   // For Boolean values: "type": 1
                        #   // For Integer values: "type": 2
                        #   // For Double values: "type": 3
                        #   // For Array values: "type": 4
                        #   "type": Integer,
                        #
                        #   // Populate with a String value if `type` is 0, otherwise do not include this field.
                        #   "string_value": String,
                        #   // Populate with a Boolean value if `type` is 1, otherwise do not include this field.
                        #   "bool_value": Boolean,
                        #   // Populate with a Integer value if `type` is 2, otherwise do not include this field.
                        #   "int_value": Integer,
                        #   // Populate with a Double value if `type` is 3, otherwise do not include this field.
                        #   "double_value": Double,
                        #   // Populate with a Array value if `type` is 4, otherwise do not include this field.
                        #   "array_value": Array<AttributeArrayValue>,
                        # }
                        # assert its values

                        if attr["type"] == 0:
                            assert isinstance(
                                attr["string_value"], str
                            ), f"Expected 'string_value' to be of type: 'str', got: {type(attr['string_value'])}"
                        elif attr["type"] == 1:
                            assert isinstance(
                                attr["bool_value"], bool
                            ), f"Expected 'bool_value' to be of type: 'bool', got: {type(attr['bool_value'])}"
                        elif attr["type"] == 2:
                            assert isinstance(
                                attr["int_value"], int
                            ), f"Expected 'int_value' to be of type: 'int', got: {type(attr['int_value'])}"
                        elif attr["type"] == 3:
                            assert isinstance(
                                attr["double_value"], float
                            ), f"Expected 'double_value' to be of type: 'float', got: {type(attr['double_value'])}"
                        elif attr["type"] == 4:
                            assert isinstance(
                                attr["array_value"], list
                            ), f"Expected 'array_value' to be of type: 'list', got: {type(attr['array_value'])}"
                            array = attr["array_value"]
                            if array:
                                first_type = array[0]["type"]
                                i = None
                                assert all(
                                    i["type"] == first_type for i in array
                                ), f"Expected all elements in list to be of the same type: '{first_type}', got: {i['type']}"

                            for e in array:
                                assert isinstance(
                                    e, dict
                                ), f"Expected all elements in 'array_value' to be of type: 'dict', got: {type(e)}"
                                if e["type"] == 0:
                                    assert isinstance(
                                        e["string_value"], str
                                    ), f"Expected 'string_value' to be of type: 'str', got: {type(e['string_value'])}"
                                elif e["type"] == 1:
                                    assert isinstance(
                                        e["bool_value"], bool
                                    ), f"Expected 'bool_value' to be of type: 'bool', got: {type(e['bool_value'])}"
                                elif e["type"] == 2:
                                    assert isinstance(
                                        e["int_value"], int
                                    ), f"Expected 'int_value' to be of type: 'int', got: {type(e['int_value'])}"
                                elif e["type"] == 3:
                                    assert isinstance(
                                        e["double_value"], float
                                    ), f"Expected 'double_value' to be of type: 'float', got: {type(e['double_value'])}"
                                else:
                                    raise ValueError(
                                        f"Unsupported span event attribute type {attr['type']} for: {k}={attr}"
                                    )
                        else:
                            raise ValueError(f"Unsupported span event attribute type {attr['type']} for: {k}={attr}")

        return cast(Span, d)
    except AssertionError as e:
        raise TypeError(*e.args) from e


def _parse_meta_struct(value: Any) -> Dict[str, Dict[str, Any]]:
    if not isinstance(value, dict):
        raise TypeError("Expected meta_struct to be of type: 'dict', got: %s" % type(value))

    return {key: msgpack.unpackb(val_bytes) for key, val_bytes in value.items()}


def _flexible_decode_meta_struct(value: Any) -> None:
    if not isinstance(value, list):
        return
    for maybe_trace in value:
        if not isinstance(maybe_trace, list):
            continue
        for maybe_span in maybe_trace:
            if not isinstance(maybe_span, dict):
                continue
            if "meta_struct" in maybe_span:
                maybe_span["meta_struct"] = _parse_meta_struct(maybe_span["meta_struct"])


def v04_verify_trace(maybe_trace: Any) -> Trace:
    if not isinstance(maybe_trace, list):
        raise TypeError("Trace must be a list.")
    for maybe_span in maybe_trace:
        verify_span(maybe_span)
    return cast(Trace, maybe_trace)


def _verify_v04_payload(data: Any) -> v04TracePayload:
    if not isinstance(data, list):
        raise TypeError("Trace payload must be a list.")
    for maybe_trace in data:
        v04_verify_trace(maybe_trace)
    return cast(v04TracePayload, data)


def child_map(trace: Trace) -> Dict[int, List[Span]]:
    cmap: Dict[SpanId, List[Span]] = {}
    # Initialize the map with all possible ids
    for s in trace:
        cmap[s["span_id"]] = []
        parent_id = s.get("parent_id") or 0
        cmap[parent_id] = []

    for s in trace:
        parent_id = s.get("parent_id") or 0
        cmap[parent_id].append(s)

    # Sort the children ascending by their start time, else by their span_id
    for span_id in cmap:
        cmap[span_id] = sorted(cmap[span_id], key=lambda _: (_["start"] if "start" in _ else _["span_id"]))
    return cmap


def bfs_order(trace: Trace) -> Generator[Span, None, None]:
    """Return trace in BFS order.

    Note: does not return copies of the spans.
    """
    cmap = child_map(trace)
    root = root_span(trace)
    children = [[root]]
    while children:
        cs = children.pop(0)
        for c in cs:
            yield c
            children.append(cmap[c["span_id"]])


def dfs_order(trace: Trace) -> Generator[Span, None, None]:
    """Return the trace in DFS order.

    Note: does not return copies of the spans.
    """
    cmap = child_map(trace)
    root = root_span(trace)
    children = [root]
    while children:
        c = children.pop(0)
        yield c
        children = cmap[c["span_id"]] + children


def dfs_order_with_depth(trace: Trace) -> Generator[Tuple[Span, int], None, None]:
    cmap = child_map(trace)
    root = root_span(trace)
    children = [(root, 0)]
    while children:
        c, depth = children.pop(0)
        yield c, depth
        children = [(_, depth) for _ in cmap[c["span_id"]]] + children


def pprint_trace(
    trace: Trace,
    fmt: Union[str, Callable[[Span], str]],
) -> str:
    cmap = child_map(trace)
    stack: List[Tuple[str, str, Span]] = [("", "", root_span(trace))]
    s = ""
    while stack:
        prefix, childprefix, span = stack.pop(0)
        for i, child in enumerate(reversed(cmap[span["span_id"]])):
            if i == 0:
                stack.insert(0, (childprefix + "└─ ", childprefix + "   ", child))
            else:
                stack.insert(0, (childprefix + "├─ ", childprefix + "│  ", child))

        spanf = fmt(span) if callable(fmt) else fmt.format(**span)
        s += f"{prefix}{spanf}"
        if stack:
            s += "\n"
    return s


def copy_span_links(s: SpanLink) -> SpanLink:
    attributes = s["attributes"].copy() if "attributes" in s else None
    copy = s.copy()
    if attributes is not None:
        copy["attributes"] = attributes
    return copy


def copy_span_events(s: SpanEvent) -> SpanEvent:
    attributes = s["attributes"].copy() if "attributes" in s else None
    copy = s.copy()
    if attributes is not None:
        # Copy arrays inside attributes
        for k, v in attributes.items():
            if isinstance(v, dict) and v["type"] == "array_value":
                array = v["array_value"]

                value = v.copy()
                value["array_value"] = array.copy()

                attributes[k] = value
        copy["attributes"] = attributes
    return copy


def copy_span(s: Span) -> Span:
    meta = s["meta"].copy() if "meta" in s else None
    metrics = s["metrics"].copy() if "metrics" in s else None
    links = s["span_links"].copy() if "span_links" in s else None
    events = s["span_events"].copy() if "span_events" in s else None
    copy = s.copy()
    if meta is not None:
        copy["meta"] = meta
    if metrics is not None:
        copy["metrics"] = metrics
    if links is not None:
        copy["span_links"] = [copy_span_links(link) for link in links]
    if events is not None:
        copy["span_events"] = [copy_span_events(event) for event in events]
    return copy


def copy_trace(t: Trace) -> Trace:
    return [copy_span(s) for s in t]


def root_span(t: Trace) -> Span:
    """Return the root span of the trace."""
    # Follow approach used in Datadog Agent: https://github.com/DataDog/datadog-agent/blob/927f9ca9acf7983b72a4bfbdd7a69132e1da8501/pkg/trace/traceutil/trace.go#L53

    if len(t) == 0:
        raise ValueError("empty trace: %s" % t)

    # common case optimization to check for span where parent_id is either not
    # set or set to 0
    for s in t:
        if "parent_id" not in s or s["parent_id"] is None or s["parent_id"] == 0:
            return s

    # collect root spans as those with parents that are not themselves spans in trace
    span_ids = set(s["span_id"] for s in t)
    roots = {s["parent_id"]: s for s in t if "parent_id" in s and s["parent_id"] not in span_ids}

    if len(roots) != 1:
        raise ValueError("single root span not found in trace (n=%d): %s" % (len(t), t))

    # return any root candidate
    return roots.popitem()[1]


def trace_id(t: Trace) -> TraceId:
    return t[0]["trace_id"]


def set_attr(s: Span, k: SpanAttr, v: TopLevelSpanValue) -> Span:
    s[k] = v
    return s


def set_meta_tag(s: Span, k: str, v: str) -> Span:
    s["meta"][k] = v
    return s


def set_metric_tag(s: Span, k: str, v: MetricType) -> Span:
    s["metrics"][k] = v
    return s


def add_span_link(
    s: Span, link: Span, attributes: Optional[Dict[str, str]] = None, flags: Optional[int] = None
) -> Span:
    if "span_links" not in s:
        s["span_links"] = []
    new_link = SpanLink(trace_id=link["trace_id"], span_id=link["span_id"], trace_id_high=0)
    if attributes is not None:
        new_link["attributes"] = attributes
    if flags is not None:
        new_link["flags"] = flags
    s["span_links"].append(new_link)
    return s


def add_span_event(
    s: Span,
    time_unix_nano: int = 1730405656000000000,
    name: str = "event",
    attributes: Optional[Dict[str, Dict[str, Any]]] = None,
) -> Span:
    if "span_events" not in s:
        s["span_events"] = []
    new_event = SpanEvent(time_unix_nano=time_unix_nano, name=name)
    if attributes is not None:
        # Add protobuf-like structure for attributes
        new_attributes: Dict[str, Dict[str, Any]] = {}
        for k, v in attributes.items():
            if isinstance(v, str):
                new_attributes[k] = {"type": 0, "string_value": v}
            elif isinstance(v, bool):
                new_attributes[k] = {"type": 1, "bool_value": v}
            elif isinstance(v, int):
                new_attributes[k] = {"type": 2, "int_value": v}
            elif isinstance(v, float):
                new_attributes[k] = {"type": 3, "double_value": v}
            elif isinstance(v, list):
                array_value: List[Dict[str, Any]] = []
                new_attributes[k] = {"type": 4, "array_value": array_value}
                for i in v:
                    if isinstance(i, str):
                        array_value.append({"type": 0, "string_value": i})
                    elif isinstance(i, bool):
                        array_value.append({"type": 1, "bool_value": i})
                    elif isinstance(i, int):
                        array_value.append({"type": 2, "int_value": i})
                    elif isinstance(i, float):
                        array_value.append({"type": 3, "double_value": i})
                    else:
                        raise ValueError(f"Unsupported span event attribute type {type(i)} for: {k}={v}")
            else:
                raise ValueError(f"Unsupported span event attribute type {type(v)} for: {k}={v}")
        new_event["attributes"] = new_attributes
    s["span_events"].append(new_event)
    return s


def _trace_decoder_flexible(json_string: bytes) -> Dict[str, Any]:
    """Parse Trace JSON and accounts for meta that may contain numbers such as ports. Converts these meta correctly to strings.
    Also ensures that any valid integers/floats are correctly parsed, to prevent ids from being decoded as strings incorrectly.
    """

    def is_number_as_str(num, number_type=int):
        try:
            number_type(num)
            return isinstance(num, str)
        except ValueError:
            return False

    # Define a custom JSON decoder for decoding spans
    def json_decoder(maybe_span):
        # loop through the span object
        if isinstance(maybe_span, dict):
            for key, value in maybe_span.items():
                if key == "meta":
                    # Check if the value is an int or float and convert back to string if true
                    for k, v in value.items():
                        if isinstance(v, int) or isinstance(v, float):
                            value[k] = str(v)
                elif key == "metrics":
                    for k, v in value.items():
                        # Check if value is a a float or int
                        if is_number_as_str(v, int):
                            value[k] = int(v)
                        elif is_number_as_str(v, float):
                            value[k] = float(v)
                # For other attributes, check if the value is a string that can be converted to an int
                elif is_number_as_str(value, int):
                    maybe_span[key] = int(value)
        return maybe_span

    parsed_data: Dict[str, Any] = json.loads(json_string, object_hook=json_decoder)
    return parsed_data


def decode_v04(content_type: str, data: bytes, suppress_errors: bool) -> v04TracePayload:
    if content_type == "application/msgpack":
        payload = msgpack.unpackb(data)
    elif content_type == "application/json":
        payload = _trace_decoder_flexible(data) if suppress_errors else json.loads(data)
    else:
        raise TypeError("Content type %r not supported" % content_type)
    _flexible_decode_meta_struct(payload)
    return _verify_v04_payload(payload)


def decode_v05(data: bytes) -> v04TracePayload:
    payload = msgpack.unpackb(data, strict_map_key=False)
    if not isinstance(payload, list):
        raise TypeError("Trace payload must be an array containing two elements, got type %r." % type(payload))
    if len(payload) != 2:
        raise TypeError("Trace payload must contain two elements, got an array with %r elements." % len(payload))

    maybe_string_table = payload[0]
    for s in maybe_string_table:
        if not isinstance(s, str):
            raise TypeError("String table contains non-string value %r." % s)
    string_table = cast(List[str], maybe_string_table)

    v05_traces = payload[1]
    traces: List[List[Span]] = []
    for v05_trace in v05_traces:
        trace: List[Span] = []
        for v05_span in v05_trace:
            if not isinstance(v05_span, list):
                raise TypeError("Span data was not an array, got type %r." % type(v05_span))
            if len(v05_span) != 12:
                raise TypeError("Span data was not an array of size 12, got array of size %r" % len(v05_span))

            v05_meta = v05_span[9]
            meta: Dict[str, str] = {}
            for idx1, idx2 in v05_meta.items():
                meta[string_table[idx1]] = string_table[idx2]

            v05_metrics = v05_span[10]
            metrics: Dict[str, MetricType] = {}
            for idx, val in v05_metrics.items():
                if not isinstance(val, (float, int)):
                    raise TypeError("Unexpected metric type %s" % type(val))
                metrics[string_table[idx]] = val

            # Recreate the span using the string values from the table
            span = verify_span(
                {
                    "service": string_table[v05_span[0]],
                    "name": string_table[v05_span[1]],
                    "resource": string_table[v05_span[2]],
                    "trace_id": v05_span[3],
                    "span_id": v05_span[4],
                    "parent_id": v05_span[5],
                    "start": v05_span[6],
                    "duration": v05_span[7],
                    "error": v05_span[8],
                    "meta": meta,
                    "metrics": metrics,
                    "type": string_table[v05_span[11]],
                }
            )
            trace.append(span)
        traces.append(trace)
    return traces


def decode_v07(data: bytes) -> v04TracePayload:
    """Decode a v07 trace payload.
    The v07 format is almost the same as the v04 format, but the there is some
    extra structure to the payload.

    These are the types of the payload:

    class TracerPayloadV07(TypedDict):
        container_id: NotRequired[str]
        language_name: NotRequired[str]
        language_version: NotRequired[str]
        tracer_version: NotRequired[str]
        runtime_id: NotRequired[str]
        chunks: List[TraceChunkV07]
        tags: NotRequired[Dict[str, str]]
        env: NotRequired[str]
        hostname: NotRequired[str]
        app_version: NotRequired[str]

    class TraceChunkV07(TypedDict):
        priority: int
        origin: str
        spans: List[Span]
        tags: NotRequired[Dict[str, str]]
        droppedTrace: NotRequired[bool]
    """
    payload = msgpack.unpackb(data)
    return _verify_v07_payload(payload)


def _verify_v07_payload(data: Any) -> v04TracePayload:
    if not isinstance(data, dict):
        raise TypeError("Trace payload must be a map, got type %r." % type(data))
    if "chunks" not in data:
        raise TypeError("Trace payload must contain a 'chunks' key.")
    if not isinstance(data["chunks"], list):
        raise TypeError("Trace payload 'chunks' must be a list.")
    # TODO:ban pull out the tags and other things that should be applied to all spans
    traces: List[List[Span]] = []
    for chunk in data["chunks"]:
        traces.append(_verify_v07_chunk(chunk))
    return cast(v04TracePayload, traces)


def _verify_v07_chunk(chunk: Any) -> List[Span]:
    if not isinstance(chunk, dict):
        raise TypeError("Chunk must be a map.")
    if "spans" not in chunk:
        raise TypeError("Chunk must contain a 'spans' key.")
    if not isinstance(chunk["spans"], list):
        raise TypeError("Chunk 'spans' must be a list.")
    # TODO:ban pull out the tags and other things that should be applied to all spans
    return v04_verify_trace(chunk["spans"])
